from schema import SchemaQP
from anndata import AnnData
import numpy as np
import scanpy as sc

from .process import load_names

def load_meta(fname):
    age, strain = [], []
    with open(fname) as f:
        f.readline() # Consume header.
        for line in f:
            fields = line.rstrip().split()
            age.append(int(fields[4]))
            strain.append(fields[3])
    return np.array(age), np.array(strain)

if __name__ == '__main__':
    [ X ], [ genes ], _ = load_names([ 'data/fly_brain/GSE107451' ], norm=False)

    age, strain = load_meta('data/fly_brain/GSE107451/annotation.tsv')

    # Only analyze wild-type strain.
    adata = AnnData(X[strain == 'DGRP-551'])
    adata.var['gene_symbols'] = genes
    adata.obs['age'] = age[strain == 'DGRP-551']

    # No Schema transformation.

    sc.pp.pca(adata)
    sc.tl.tsne(adata, n_pcs=50)
    sc.pl.tsne(adata, color='age', color_map='coolwarm',
               save='_flybrain_regular.png')

    sc.pp.neighbors(adata, n_neighbors=15)
    sc.tl.umap(adata)
    sc.pl.umap(adata, color='age', color_map='coolwarm',
               save='_flybrain_regular.png')

    # Schema transformation to include age.

    schema_corrs = [ 0.9999, 0.999, 0.99, 0.9, 0.7, 0.5 ]

    for schema_corr in schema_corrs:

        sqp = SchemaQP(
            min_desired_corr=schema_corr,
            w_max_to_avg=100,
            params={
                'decomposition_model': 'nmf',
                'num_top_components': 20,
            },
        )

        X = sqp.fit_transform(
            adata.X,
            [ adata.obs['age'].values, ],
            [ 'numeric', ],
            [ 1, ]
        )

        sdata = AnnData(X)
        sdata.obs['age'] = age[strain == 'DGRP-551']

        sc.tl.tsne(sdata)
        sc.pl.tsne(sdata, color='age', color_map='coolwarm',
                   save='_flybrain_schema_corr{}_w100.png'.format(schema_corr))

        sc.pp.neighbors(sdata, n_neighbors=15)
        sc.tl.umap(sdata)
        sc.pl.umap(sdata, color='age', color_map='coolwarm',
                   save='_flybrain_schema{}_w100.png'.format(schema_corr))



#!/usr/bin/env python

import logging

schema_loglevel = logging.WARNING #WARNING


def schema_debug(*args, **kwargs):
    if schema_loglevel <= logging.DEBUG: print("DEBUG: ", *args, **kwargs)

def schema_info(*args, **kwargs):
    if schema_loglevel <= logging.INFO: print("INFO: ", *args, **kwargs)

def schema_warning(*args, **kwargs):
    if schema_loglevel <= logging.WARNING: print("WARNING: ", *args, **kwargs)

def schema_error(*args, **kwargs):
    if schema_loglevel <= logging.ERROR: print("ERROR: ", *args, **kwargs)

    
########## for maintenance ###################
# def noop(*args, **kwargs):
#     pass
#
# logging.info = print
# logging.debug = noop
##############################################



#!/usr/bin/env python

import pandas as pd
import numpy as np
import scipy, sklearn, os, sys, string, fileinput, glob, re, math, itertools, functools, copy, multiprocessing, traceback
import scipy.stats, sklearn.decomposition, sklearn.preprocessing, sklearn.covariance
from scipy.stats import describe
from scipy import sparse
import os.path
import scipy.sparse
from scipy.sparse import csr_matrix, csc_matrix
from sklearn.preprocessing import normalize
from collections import defaultdict
from tqdm import tqdm



def read_raw_files_and_write_h5_files(outdir):
    def f_atac1(w):
        try:
            v = w["chr"]
            assert ("_" not in v) and (v in ['X','Y'] or int(v)<=22) 
            return True
        except:
            return False

    import utils
    adata1 = utils.SciCar.loadData('/afs/csail.mit.edu/u/r/rsingh/work/schema/data/sci-car',
                                [('rna' ,'gene','GSM3271040', lambda v: v["cell_name"]=="A549", lambda v: v["gene_type"]=="protein_coding"),
                                 ('atac','peak','GSM3271041', lambda v: v["group"][:4]=="A549", f_atac1),
                                ],
                                   "/afs/csail.mit.edu/u/r/rsingh/work/refdata/hg19_mapped.tsv")
    adata1.write("{0}/adata1x.h5ad".format(outdir))

    
def mprun_geneXfpeak_mtx(adata, n_jobs=8):
    peak_func_list = [a[0] for a in SciCar.fpeak_list_all]
    print (adata.uns.keys(), adata.uns["atac.var"].head(2))
    chr_mapping = SciCar.getChrMapping(adata)


    nPeaks = adata.uns["atac.X"].shape[1]
    l = [list(a) for a in np.array_split(range(nPeaks), 5*n_jobs)]
    pool =  multiprocessing.Pool(processes = n_jobs)

    lx = pool.map(functools.partial(SciCar.computeGeneByFpeakMatrix, 
                                    adata, peak_func_list, chr_mapping, normalize_distwt= True),
                  l)


    g2p = None
    for m, _ in lx:
        if g2p is None:
            g2p = m
        else:
            g2p += m

    g2p =  g2p * (1e5/nPeaks)

    dx = pd.DataFrame(g2p, index=None)
    dx.columns = [a[1] for a in SciCar.fpeak_list_all]
    dx["gene"] = list(adata.var.index)
    dx["ensembl_id"] = list(adata.var.ensembl_id)               
    return dx




def f_helper_mprun_schemawts_2(args):
    ax, dz, dir1, outsfx, min_corr, maxwt, strand, chromosome, adata_norm_style = args

    if adata_norm_style == 1:
        ax_l2norm = np.sqrt(np.sum(ax**2, axis=1))
        ax = ax.copy() / (1e-12 + ax_l2norm[:,None])
        
    elif adata_norm_style == 2:
        ax_l2norm = np.sqrt(np.sum(ax**2, axis=1))
        ax = np.sort(ax.copy(), axis=1) / (1e-12 + ax_l2norm[:,None])
        print("Flag 231.10 ")
        print(ax[-10:,-10:])
        print(np.sum(ax**2,axis=1)[-10:])
        
    dz_cols = dz.columns[:-2]
    dz_vals = dz.values[:,:-2]
        
    vstd = np.std(dz_vals.astype(float), axis=0)
    print("Flag 231.0275 ", vstd.shape, dz_vals.shape, flush=True)
    dz_vals = dz_vals.copy() / (1e-12 + vstd)
            
    try:
        sys.path.append(os.path.join(sys.path[0],'../../schema'))
        import schema_qp
    except:
        from schema import schema_qp

    sqp = schema_qp.SchemaQP(min_corr, maxwt, params= {"dist_npairs": 1000000}, mode="scale")
    try:
        dz1 = sqp.fit_transform(dz_vals, [ax], ['feature_vector'], [1])
        
        print("Flag 231.030 ", min_corr, maxwt, dz_vals.shape, ax.shape, flush=True)
    
        wtsx = np.sqrt(np.maximum(sqp._wts/np.sum(sqp._wts), 0))
    except:
        print ("ERROR: schema failed for ", min_corr, maxwt, strand, chromosome)
        wtsx = 1e12*np.ones(dz_vals.shape[1])
    
    wdf = pd.Series(wtsx, index=dz_cols).sort_values(ascending=False).reset_index().rename(columns={"index": "fdist",0: "wt"})
    wdf.to_csv("{0}/adata1_sqp_wts_mincorr{1}_maxw{2}_strand{4}_chr{5}_adatanorm{8}_{7}.csv".format(dir1, min_corr, maxwt, 1, strand, chromosome, 5, outsfx, adata_norm_style), index=False)


    
def mprun_schemawts_2(adata1, dz, dir1, outsfx, adata_norm_style, do_dataset_split=False, n_jobs=4):
    
    try:
        sys.path.append(os.path.join(sys.path[0],'../../schema'))
        import schema_qp
    except:
        from schema import schema_qp
    
    pool =  multiprocessing.Pool(processes = n_jobs)
    try:
        ax = np.copy(adata1.X.todense().T)
    except:
        ax = adata1.X.T.copy()

    nGenes = adata1.var.shape[0]
    chrx = adata1.var["chr"].apply(lambda s: s.replace("chr",""))
    
    lx = []

    # one-time export of data for ridge-regression comparison
    # if False:
    #     dfax = pd.DataFrame(ax.T)
    #     dfax.columns = list(adata1.var_names)
        
    #     dz_vals = dz[dz.ensembl_id.isin(adata1.var["ensembl_id"])].values.copy()[:,:-2]
    #     vstd = np.std(dz_vals.astype(float), axis=0)
    #     dfdz = pd.DataFrame(dz_vals/ (1e-12 + vstd))
    #     dfdz.columns = list(dz.columns)[:-2]
    #     dfax.to_csv("saved_data_{0}_gexp.csv".format(outsfx), index=False)
    #     adata1.var.to_csv("saved_data_{0}_aux.csv".format(outsfx), index=False)
    #     dfdz.to_csv("saved_data_{0}_features.csv".format(outsfx), index=False)
    

    dz2 = dz
    for strand in ["both", "plus", "minus"]:
        if strand != "both" and do_dataset_split==False: continue
        
        gidx_strand = np.full(nGenes, True)
        if strand == "plus":  gidx_strand = np.where(adata1.var["strand"]=="+", True, False)
        if strand == "minus": gidx_strand = np.where(adata1.var["strand"]=="-", True, False)

        for chromosome in ["all", "1--8","9--16","17--23"]:
            if chromosome != "all" and do_dataset_split==False: continue

            gidx_chr = np.full(nGenes, True)
            if chromosome=="1--8":   gidx_chr = chrx.isin("1,2,3,4,5,6,7,8".split(","))
            if chromosome=="9--16":  gidx_chr = chrx.isin("9,10,11,12,13,14,15,16".split(","))
            if chromosome=="17--23": gidx_chr = chrx.isin("17,18,19,20,21,22,X,Y".split(","))

            for mc in [0.20, 0.5, 0.9]:
                for mw in [10,5,3]:

                    gidx = gidx_strand & gidx_chr
                    print("Flag 3312.040 ", gidx.shape, np.sum(gidx), ax.shape, dz2.shape, flush=True)
                    lx.append((ax[gidx,:], dz2[dz2.ensembl_id.isin(adata1.var["ensembl_id"][gidx])], dir1, outsfx,
                               mc, mw, strand, chromosome, adata_norm_style))
                    print("Flag 3312.050 ", np.sum(gidx), lx[-1][2:], flush=True)

    try:
        pool.map(f_helper_mprun_schemawts_2, lx)

    finally:
        pool.close()
        pool.join()




def  getGeneDistances(adata1):
    chrInt = adata1.var["chr"].apply(lambda s: s.replace("chr","").replace("X","231").replace("Y","232")).astype(int).values
    diff_chr = np.subtract.outer(chrInt, chrInt)!=0

    tss_adj = adata1.var['tss_adj'].values
    chrdist = np.abs(np.subtract.outer(tss_adj, tss_adj))

    import numpy.ma as ma
    gene_dist = ma.masked_where( diff_chr, chrdist).filled(1e15)
    nGenes = gene_dist.shape[0]
    l0 = list(range(nGenes))
    gene_dist[l0, l0] = 1e15
    return gene_dist


def  getGeneStrandMatch(adata1):
    chrInt = adata1.var["chr"].apply(lambda s: s.replace("chr","").replace("X","231").replace("Y","232")).astype(int).values
    diff_chr = np.subtract.outer(chrInt, chrInt)!=0

    strand_adj = np.where(adata1.var['strand'].values=="+",1,0)
    strandM = (np.subtract.outer(strand_adj, strand_adj) == 0).astype(int)

    import numpy.ma as ma
    gene_sign = ma.masked_where( diff_chr, strandM).filled(1e15)
    nGenes = gene_sign.shape[0]
    l0 = list(range(nGenes))
    gene_sign[l0, l0] = 1e15
    return gene_sign


def rankGenesByVariability(adata1):
    import matplotlib
    matplotlib.use("agg") #otherwise scanpy tries to use tkinter which has issues importing
    import scanpy as sc 
    nGenes = adata1.shape[1]
    print ("Flag 7868.200 ",   nGenes, adata1.X.shape, adata1.shape, scipy.sparse.isspmatrix(adata1.X))
    v = np.full(nGenes, nGenes)
    for i in range(500,nGenes-500,250):
        print ("Flag 7868.205 ",   i)
        
        v0 = sc.pp.highly_variable_genes(adata1, n_top_genes=i,inplace=False).highly_variable
        print ("Flag 7868.210 ",   i, np.sum(v0), len(v0))
        v = np.where((v==nGenes) & v0, i, v).copy()
    return v


def findNonVariableGenes(adata1, gene_window):
    nGenes = adata1.shape[1]

    v = np.full(nGenes, nGenes)
    for i in range(250,nGenes-500,250):
        try:
            hv = sc.pp.highly_variable_genes(adata1, n_top_genes=i,inplace=False).highly_variable
        except:
            print ("Flag 7867.500 ", i)
            traceback.print_exc(file=sys.stdout)
            continue

        v = np.where((v==nGenes) & hv, i, v).copy()


    w = (v  >= gene_window[0]) & (v < gene_window[1])
    print ("Flag 7867.200 ",  np.sum(w), gene_window, nGenes)
    print ("Flag 7867.300 ",  pd.Series(v).value_counts())

    return w



def random_gene_list(nGenes, gene_cnt, num_random_samples=1000):
    L  = []
    for i in range(num_random_samples):
        b = np.zeros(nGenes, dtype=bool)
        b[np.random.choice(nGenes, gene_cnt, replace=False)] = True
        L.append(b)

    return L
    

def  getGene2TADmap(adata1, tad):
    l = []
    nGenes = adata1.var.shape[0]
    for i in tqdm(range(nGenes)):
        gchr = adata1.var["chr"][i]
        gtss = adata1.var["tss_adj"][i]
        idx1 = (tad.chr==gchr) & (tad["start"] <= gtss) & (tad["end"] > gtss)
        #if i%100==0: print ("Flag 3231.10 ", i)
        if idx1.sum() > 0:
            l.append( np.argmax(idx1.values))
        else:
            l.append(np.NaN)
    g2tad = np.array(l)
    gene_sharedtad = np.subtract.outer( g2tad, g2tad)==0
    return g2tad, gene_sharedtad



def f_in_tad_cnt( g2tad, gene_set):
    return np.sum(~np.isnan(g2tad[gene_set]))


def f_mode_hvg_tad_dispersion_helper(g2tad, gene_sharedtad,  gene_set):
    tad2genes = defaultdict(list)
    for i, t in enumerate(g2tad):
        if not np.isnan(t) and gene_set[i]:
            tad2genes[t].append(i)
    return [(k,v,len(v)) for k,v in tad2genes.items()]


    
#################################################################################

if __name__ == "__main__":
    try:
        sys.path.append(os.path.join(sys.path[0],'../../schema'))
        from utils import SciCar
    except:
        import schema
        from schema.utils import SciCar

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", help="which code path to run. see main(..) for details")
    parser.add_argument("--outdir", help="output directory (can set to '.')", type=str, default=".")
    parser.add_argument("--outsfx", help="suffix to use when producing output files")
    parser.add_argument("--style", help="mode-specific interpretation", type=int, default=-1)
    parser.add_argument("--infile", help="input .h5ad file. Default is Sci-Car hs A549")
    parser.add_argument("--njobs", help="number of parallel cores to use", type=int, default=24)

    parser.add_argument("--extra", help="put this as the LAST option and arbitrary space-separated key=val pairs after that", type=str, nargs='*')

    
    args = parser.parse_args()
    assert args.mode is not None
    if args.mode !="raw_data_read": assert args.outsfx is not None

    extra_args = dict([a.split("=") for a in args.extra]) if args.extra else {}

    if args.mode== "raw_data_read":
        read_raw_files_and_write_h5_files( args.outdir)



    if args.infile is None:
        args.infile = "/afs/csail.mit.edu/u/r/rsingh/work/schema/data/sci-car/processed/adata1x.h5ad"


    adata1 = SciCar.loadAnnData(args.infile)
    adata1 = SciCar.preprocessAnnData(adata1, True, 5, 3, 5)

    if args.mode=="produce_gene2fpeak":
        if args.style < 0: args.style = 0
        dx = mprun_geneXfpeak_mtx(adata1, 36)
        dx.to_csv("{0}/adata1_M-{1}_S-{2}_mtx_{3}.csv".format( args.outdir, args.mode, args.style, args.outsfx), index=False)
        
        
    if args.mode=="schema_gene2fpeak":
        #  --mode schema_gene2fpeak --outsfx 20191218-1615 --njobs 5 --extra gene2fpeak_file=adata1_M-produce_gene2fpeak_S-0_mtx_20191218-1615.csv

        if args.style < 0: args.style = 0

        assert "gene2fpeak_file" in extra_args
        dz = pd.read_csv(extra_args["gene2fpeak_file"])

        if "fpeak_cols_to_drop" in extra_args:
            dz = dz.drop(columns = extra_args["fpeak_cols_to_drop"].split(","))

        adata1.var["gene_variability_ranking"] = rankGenesByVariability(adata1)
        
        adata1.X = adata1.X.todense()
        
        min_hvgrank, max_hvgrank = int(extra_args.get("min_hvgrank",0)), int(extra_args.get("max_hvgrank",1000000))
        adata1 = adata1[:, ((adata1.var["gene_variability_ranking"] >= min_hvgrank) & 
                            (adata1.var["gene_variability_ranking"] < max_hvgrank))]

        dz = dz[dz.ensembl_id.isin(adata1.var["ensembl_id"])].reset_index(drop=True)

        assert np.sum(dz.ensembl_id.values != adata1.var["ensembl_id"].values)==0

        adata_norm_style = int(extra_args.get("adata_norm_style",0))
        do_dataset_split = int(extra_args.get("do_dataset_split",0)) > 0.5
        
        mprun_schemawts_2(adata1, dz, args.outdir, "M-{0}_S-{1}_minhvg-{2}_maxhvg-{3}_{4}".format(args.mode, args.style, min_hvgrank, max_hvgrank, args.outsfx), adata_norm_style, do_dataset_split, args.njobs)



    if args.mode == "compute_hvg_tad_dispersion": 
        import scanpy as sc

        min_hvgrank, max_hvgrank = int(extra_args.get("min_hvgrank",0)), int(extra_args.get("max_hvgrank",1000000))
        gene_window = [min_hvgrank, max_hvgrank]

        outsfx = args.outsfx

        gdist_matrix = getGeneDistances(adata1)
        print ("Flag 54123.20 ", gdist_matrix.shape) #, describe(gdist_matrix))

        gsign_matrix = getGeneStrandMatch(adata1)
        print ("Flag 54123.21 ", gsign_matrix.shape) #, describe(gdist_matrix))

        tad = pd.read_csv(extra_args.get("tad_locations_file", "hg19_A549_TAD.bed"), delimiter="\t", header=None)
        tad.columns = ["chr","start","end", "x1","x2"]
        g2tad, gene_sharedtad = getGene2TADmap(adata1, tad)

        nGenes = adata1.shape[1]
        
        # v = np.full(nGenes, nGenes)
        # for i in tqdm(range(500,nGenes-500,500)):
        #     v0 = sc.pp.highly_variable_genes(adata1, n_top_genes=i,inplace=False).highly_variable
        #     v = np.where((v==nGenes) & v0, i, v).copy()
        # adata1.var["gene_variability_ranking"] = v

        adata1.var["gene_variability_ranking"] = rankGenesByVariability(adata1)
            
        window_genes = ((adata1.var["gene_variability_ranking"].values >= gene_window[0]) &  
                        (adata1.var["gene_variability_ranking"].values <   gene_window[1])) 

        n_window_genes = np.sum(window_genes)

        print ("Flag 54123.30 ", gene_window, len(window_genes), np.sum(window_genes))

        num_samples = 1000

        in_tad_cnt = f_in_tad_cnt(g2tad, window_genes)
        tad_total_cnt = np.sum(~np.isnan(g2tad))

        rl = random_gene_list(tad_total_cnt, in_tad_cnt, num_random_samples = num_samples)

        # we don't really care about genes outside TADs in this analysis
        window_genes [np.isnan(g2tad)] = False
        n_window_genes = np.sum(window_genes)
        lx0 = [window_genes]
        print ("Flag 54123.32 ",  len(window_genes), np.sum(window_genes), len(lx0))
        for rlx in rl:
            ax = np.full( nGenes, False)
            ax[~np.isnan(g2tad)] = rlx #[rlx] = True
            lx0.append(ax)
        print ("Flag 54123.33 ",  len(window_genes), np.sum(window_genes), len(lx0), len(lx0[1]), np.sum(lx0[1]))
            
        lx =  [(g2tad, gene_sharedtad, a) for a in lx0]


        print ("Flag 54123.40 ", len(lx), len(lx[1][2]), np.sum(lx[1][2]))
        
        n_jobs = 36
        print ("Flag 54123.435 ", nGenes, n_window_genes, gdist_matrix.shape, adata1.shape, len(lx))


        pool =  multiprocessing.Pool(processes = n_jobs)
        ly = pool.starmap(f_mode_hvg_tad_dispersion_helper, lx)

        outfile = "adata1_tad_membership_genewindow{0}-{1}_{2}".format(min_hvgrank, max_hvgrank, outsfx)
        outfh = open(outfile, 'w')

        prep_significance_testing_data = int(extra_args.get("prep_significance_testing_data",1)) > 0.5
        
        if not prep_significance_testing_data:
            random_samples_cnts = defaultdict(int)
            random_samples_N = 0
            for z in ly[1:]:
                for _, _, tadcnt in z: 
                    random_samples_cnts[tadcnt] += 1
                random_samples_N += 1.0

            for k,v in sorted(random_samples_cnts.items(), key = lambda a: a[0]):
                outfh.write("Random,{0},{1}\n".format(k, v/random_samples_N))

            for tadidx, tad_members, tadcnt in sorted(ly[0], key=lambda a: -a[2]):
                outfh.write("Orig,{0},{1},{2}\n".format( tad["x1"][tadidx], tadcnt, ",".join([adata1.var["Symbol"][t] for t in tad_members])))

        else:
            for i , z in enumerate(ly):
                #print ("Flag 54123.50 ", i, z)
                tadcnt2freq = defaultdict(int)
                numpairs_shared_tad = 0.0
                n = 0.0
                for _, _, tadcnt in z: 
                    tadcnt2freq[tadcnt] += 1
                    numpairs_shared_tad += tadcnt*(tadcnt-1)/2.0
                    n += tadcnt
                numpairs_all = n*(n-1)/2.0
                s = "Random" if i>0 else "Orig"
                s += ",{0},{1}".format(n, numpairs_shared_tad/numpairs_all)
                s += ",".join([""]+["{0}".format(tadcnt2freq[j]) for j in range(1,16)])
                s += ",".join([""]+["{0}".format(tadcnt2freq[j]*j/n) for j in range(1,16)])
                #print("Flag 54123.56 ", s)
                outfh.write(s + "\n")
            

        outfh.close()





#!/usr/bin/env python


###################################################################
## Primary Author:  Rohit Singh rsingh@alum.mit.edu
## Co-Authors: Ashwin Narayan, Brian Hie {ashwinn,brianhie}@mit.edu
## License: MIT
## Repository:  http://github.io/rs239/schema
###################################################################

import sys, copy, os
import scanpy as sc

#### local directory imports ####
oldpath = copy.copy(sys.path)
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__))+"/../")

from schema_base_config import *

sys.path = copy.copy(oldpath)
####




def fly_brain():
    """ Anndata object containing scRNA-seq data of the ageing Drosophila brain (GSE107451, Davie et al., Cell 2018)
"""
    
    adata = sc.read("datasets/Davie_fly_brain.h5", backup_url="http://schema.csail.mit.edu/datasets/Davie_fly_brain.h5")
    return adata




#!/usr/bin/env python

import pandas as pd
import numpy as np
import scipy, sklearn, os, sys, string, fileinput, glob, re, math, itertools, functools
import  copy, multiprocessing, traceback, logging, pickle
import scipy.stats, sklearn.decomposition, sklearn.preprocessing, sklearn.covariance
from scipy.stats import describe
from scipy import sparse
import os.path
import scipy.sparse
from scipy.sparse import csr_matrix, csc_matrix
from sklearn.preprocessing import normalize
from collections import defaultdict
from tqdm import tqdm



def read_raw_files_and_write_h5_files(outdir):
    import utils
    for pid in ['180430_1','180430_5','180430_6']:
        adata1 = utils.SlideSeq.loadRawData("/afs/csail.mit.edu/u/r/rsingh/work/afid/data/slideseq/raw/", pid, 100)
        adata1.write("/afs/csail.mit.edu/u/r/rsingh/work/afid/data/slideseq/processed/puck_{0}.h5ad".format(pid))
                                     

def computeKernelDensityGranuleCells(adata1, kd_fit_granule_only=True, kd_bw=125):
    from sklearn.neighbors import KernelDensity
    fscl = lambda v: 20*(sklearn.preprocessing.MinMaxScaler().fit_transform(np.exp(v[:,None]-v.min()))).ravel()
    d3 = adata1.obs.copy(deep=True) #adata1.uns["Ho"].merge(adata1.obs, how="inner", left_index=True, right_index=True)
    d3c = d3[["xcoord","ycoord"]]
    if kd_fit_granule_only:
        d3["kd"] = fscl(KernelDensity(kernel='gaussian', bandwidth=kd_bw).fit(d3c[d3["atlas_cluster"]==1].values).score_samples(d3c.values))
    else:
        d3["kd"] = fscl(KernelDensity(kernel='gaussian', bandwidth=kd_bw).fit(d3c.values).score_samples(d3c.values))
    adata1.obs["kd"] = d3["kd"]
    return adata1



def checkMaxFeasibleCorr(D, d0, g, tg, wg):
    try:
        sys.path.append(os.path.join(sys.path[0],'../../schema'))
        import schema_qp
    except:
        from schema import schema_qp
    
    for thresh in [0.30, 0.275, 0.25, 0.225, 0.20, 0.15, 0.10, 0.075, 0.06, 0.05, 0.04, 0.03, 0.025, 0.02, 0.015, 0.01]:
        print ("STARTING TRY OF ", thresh)
        try:
            sqp = schema_qp.SchemaQP(thresh, w_max_to_avg=1000, params= {"dist_npairs": 1000000}, mode="scale")
            dz1 = sqp.fit(D, g, tg, wg, d0=d0)
            print ("SUCCEEDED TRY OF ", thresh)
            return 0.9*thresh, thresh
        except:
            print ("EXCEPTION WHEN TRYING ", thresh)
            #raise
    return 0,0


def runSchemaGranuleCellDensity(D, d0, gIn, tgIn, wgIn, min_corr1, min_corr2):
    try:
        sys.path.append(os.path.join(sys.path[0],'../../schema'))
        import schema_qp
    except:
        from schema import schema_qp
    
    f_linear = lambda v:v
    ret_val = {}

    w_list= [1,10,50,100]
    for w in w_list:
        s="linear"
        f=f_linear
        
        g1, wg1, tg1  = gIn[:], wgIn[:], tgIn[:]  # does maximize negative corr with non-granule 
        wg1[0] = w
        
        g = [g1[0]]; wg = [wg1[0]]; tg=[tg1[0]]  # does NOT maximize negative corr with non-granule
            
        #afx0 = schema_qp.SchemaQP(0.001, 1000, mode="scale")
        #Dx0 = afx0.fit_transform(D,g,tg,wg,d0)
        #ret_val[(s,w,0)] = (np.sqrt(afx0._wts), afx0._soln_info)

        try:
            afx1 = schema_qp.SchemaQP(min_corr1, w_max_to_avg=1000, mode="scale")
            Dx1 = afx1.fit_transform(D,g,tg,wg,d0=d0)
            ret_val[(s,w,1)] = (np.sqrt(afx1._wts), afx1._soln_info )  # does NOT maximize negative corr with non-granule
        except:
            print("TRYING min-corr {0} for afx1 broke here".format(min_corr1))
            continue

        try:
            afx2 = schema_qp.SchemaQP(min_corr1, w_max_to_avg=1000, mode="scale")
            Dx2 = afx2.fit_transform(D,g1,tg1,wg1,d0=d0)
            ret_val[(s,w,2)] = (np.sqrt(afx2._wts), afx2._soln_info)  # does maximize negative corr with non-granule
        except:
            print("TRYING min-corr {0} for afx2 broke here".format(min_corr1))
            continue

        try:
            afx3 = schema_qp.SchemaQP(min_corr2, w_max_to_avg=1000, mode="scale")
            Dx3 = afx3.fit_transform(D,g1,tg1,wg1,d0=d0)  # does maximize negative corr with non-granule
            ret_val[(s,w,3)] = (np.sqrt(afx3._wts), afx3._soln_info)
        except:
            print("TRYING min-corr {0} for afx3 broke here".format(min_corr2))
            continue 

    return ret_val



def getSoftmaxCombinedScores(Wo, schema_ret, use_generanks=True, do_pow2=True, schema_allowed_w1s=None): 
    if use_generanks:
        R = Wo.rank(axis=1, pct=True).values.T
    else:
        R = Wo.values.T
        
    sumr = None; nr=0
    for x in schema_ret:
        style,w,i = x

        if style!="linear" or i not in [2,3]: continue   #2,3 correspond to schema runs that also require disagreement w other altas clusters

        if schema_allowed_w1s is not None and w not in schema_allowed_w1s: continue
        
        wx = schema_ret[x][0]**2
        #wx = wx/np.sum(wx)
        
        schema_wts = wx**(2 if do_pow2 else 1)
        if np.max(schema_wts) > 20:
            schema_wts = 20*schema_wts/np.max(schema_wts)
        schema_probs = np.exp(schema_wts)/np.sum(np.exp(schema_wts))
        g1 = (R*schema_probs).sum(axis=1)
        g2 = g1/np.std(g1.ravel())
        #g2 = scipy.stats.rankdata(g1); g2 = g2/np.max(g2)
        if sumr is None:
            sumr = g2
        else:
            sumr += g2
        nr += 1
    rnks = sumr/nr
    s1= pd.Series(rnks, index=list(Wo.columns)).sort_values(ascending=False)
    return {u:(i+1) for i,u in enumerate(list(s1.index))}



def getCellLoadingSoftmax(d3, Wo, schema_ret):
    R = Wo.values.T
    
    sumr = None; nr=0
    for x in schema_ret:
        style,w,i = x
        if style!="linear" or i not in [2,3]: continue               
        
        wx = schema_ret[x][0]**2
        #wx = wx/np.sum(wx)
        
        schema_wts = wx
        if np.max(schema_wts) > 20:
            schema_wts = 20*schema_wts/np.max(schema_wts)

        schema_probs = np.exp(schema_wts)/np.sum(np.exp(schema_wts))

        if sumr is None:
            sumr = schema_probs
        else:
            sumr += schema_probs
        nr += 1
        
    r = sumr/nr
    v = (d3.iloc[:,:100].multiply(r,axis=1)).sum(axis=1)
    return v.values




def generatePlotGranuleCellDensity(d3, cell_loadings):
    import matplotlib.pyplot as plt
    import seaborn as sns
    score1 = cell_loadings

    np.random.seed(239)
    plt.style.use('seaborn-paper')
    plt.rcParams['lines.markersize'] = np.sqrt(0.25)
    fc = lambda v: np.where(v,'lightslategray','red')

    #fig = plt.figure(constrained_layout=True, figsize=(6.48,2.16), dpi=300) #(2*6.48,2*2.16))
    fig = plt.figure(figsize=(6.48,2.16), dpi=300) #(2*6.48,2*2.16))
    gs = fig.add_gridspec(2,6,wspace=0,hspace=0) #(2, 2)  

    idxY=d3["atlas_cluster"]==1
    idxN=d3["atlas_cluster"]!=1

    coords = d3.loc[:,["xcoord","ycoord"]].values
    clstr = d3["atlas_cluster"].values

    cid_list = [1,2,3,6]

    fc = lambda v: np.where(v,'lightslategray','red')
    axdict = {}

    xyL = [(gs[0,0], coords[clstr==1,:], 'a','Granule Cells'), (gs[0,1], coords[clstr==2,:],'b','Purkinje Cells'), 
           (gs[1,0], coords[clstr==3,:], 'c','Interneuron'), (gs[1,1], coords[clstr==6,:], 'd','Oligodendrocytes')]

    for g, dx, titlestr, desc in xyL:

        ax = fig.add_subplot(g)
        fc = lambda v: np.where(v,'lightslategray','red')
        ax.text(.95, .05, titlestr, horizontalalignment='center', transform=ax.transAxes, size=14 )
        ax.axis('off')
        ax.scatter(dx[:,0], dx[:,1], color='black', alpha=0.20 if titlestr=='a' else 0.6 ) 
        ax.set_aspect('equal')
        axdict[titlestr] = ax


    ax = fig.add_subplot(gs[:,2:4])
    im = ax.scatter(coords[clstr==1,0],coords[clstr==1,1],c=2*d3["kd"].values[clstr==1],cmap="seismic",s=1)
    #im = ax.scatter(coords[:,0],coords[:,1],c=2*d3["kd"].values,cmap="seismic",s=1)
    ax.set_aspect('equal')
    ax.axis('off')
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    div = make_axes_locatable(ax)
    cax = div.append_axes("bottom", size="3%", pad=0.01)
    cbar = fig.colorbar(im, cax=cax, shrink=0.2, orientation='horizontal')
    ax.text(.9, .05, "e", horizontalalignment='center', transform=ax.transAxes, size=14 )

    sx = score1 > np.quantile(score1,0.75)
    for g, titlestr, ii, c1 in [(gs[0,4], "f", idxY & sx, 'r'), (gs[0,5], "g", idxY & (~sx), 'b'), 
                                (gs[1,4], "h", idxN & sx, 'r'), (gs[1,5], "i", idxN & (~sx), 'b')]:

        ax = fig.add_subplot(g)
        ax.text(.95, .05, titlestr, horizontalalignment='center', transform=ax.transAxes, size=14 )
        #ax.axes.get_xaxis().set_visible(False)
        #ax.axes.get_yaxis().set_visible(False)
        ax.axis('off')
        ax.scatter(coords[ii,0], coords[ii,1], color=c1, alpha=0.40 ) 
        ax.set_aspect('equal')
        axdict[titlestr] = ax


    ####################################
    fig.tight_layout()
    return fig




def processGranuleCellDensitySchema(adata1, extra_args):
    
    if "kd" not in adata1.obs.columns:
        adata1 = computeKernelDensityGranuleCells(adata1, 
                                                  kd_fit_granule_only = int(extra_args.get("kd_fit_granule_only",1))==1,
                                                  kd_bw = float(extra_args.get("kd_bw",125)))

    d3 = adata1.uns["Ho"].merge(adata1.obs, how="inner", left_index=True, right_index=True)
    Wo = adata1.uns["Wo"]        
    cols_Ho = list(adata1.uns["Ho"].columns)

    D = d3[cols_Ho].values
    d0 = 1*(d3["atlas_cluster"].values==1)
    g = [(d3["kd"].values)]; wg=[10]; tg=["numeric"]
    for clid in [2,3,6,7]:
        g.append(1*(d3["atlas_cluster"].values==clid))
        wg.append(-1)
        tg.append("categorical")


    min_corr1, min_corr2 = checkMaxFeasibleCorr(D, d0, g, tg, wg)
    schema_ret = runSchemaGranuleCellDensity(D, d0, g, tg, wg, min_corr1, min_corr2)
    scores = getSoftmaxCombinedScores(Wo, schema_ret, use_generanks=False, do_pow2=False)
    cell_loadings = getCellLoadingSoftmax(d3, Wo, schema_ret)

    fig = generatePlotGranuleCellDensity(d3, cell_loadings)
    return (fig, d3, schema_ret, min_corr1, min_corr2, scores, cell_loadings)



def doSchemaCCA_CellScorePlot2(d3, cca_x_scores, cell_loadings):
    clstrs = d3["atlas_cluster"]
    kd = d3["kd"]

    cca_sgn = np.sign(scipy.stats.pearsonr(d3["kd"],cca_x_scores)[0]) #flip signs if needed

    R = {}
    for desc,v in [("ccax", cca_sgn*cca_x_scores), ("schema", cell_loadings)]:
        vr = scipy.stats.rankdata(v)
        vr = vr/vr.max()
        l = []
        for t in np.linspace(0,1,100)[:-1]:
            cx = clstrs[vr >=t ]
            granule_frac = (np.sum(cx==1)/(1e-12+ len(cx)))
            cx2 = kd[ vr >= t]
            kd_val = np.median(cx2)
            l.append((granule_frac, kd_val))
        R[desc]= list(zip(*l))

    import matplotlib.pyplot as plt
    plt.style.use('seaborn-paper')
    plt.rcParams['lines.markersize'] = np.sqrt(0.25)

    fig = plt.figure(dpi=300) #(2*6.48,2*2.16))

    a = np.linspace(0,1,100)
    plt.scatter(R["ccax"][0], R["ccax"][1], s=(1+3*a)**2, c="red", figure=fig)
    plt.scatter(R["schema"][0], R["schema"][1], s=(1+3*a)**2, c="blue", figure=fig)
    fig.legend("CCA fit,Schema fit".split(","))
    plt.xlabel("Fraction of Beads labeled as Granule Cells", figure=fig)
    plt.ylabel("Median Kernel Density Score", figure=fig)
    return fig



        
#################################################################################

if __name__ == "__main__":
    try:
        sys.path.append(os.path.join(sys.path[0],'../../schema'))
        from utils import SlideSeq
    except:
        from schema.utils import SlideSeq

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", help="which code path to run. see main(..) for details")
    parser.add_argument("--outdir", help="output directory (can set to '.')", type=str, default=".")
    parser.add_argument("--outpfx", help="prefix to use when producing output files")
    parser.add_argument("--style", help="mode-specific interpretation", type=int, default=-1)
    parser.add_argument("--infile", help="input .h5ad file. Default is SlideSeq 180430_1 h5ad")
    parser.add_argument("--njobs", help="number of parallel cores to use", type=int, default=24)

    parser.add_argument("--extra", help="put this as the LAST option and arbitrary space-separated key=val pairs after that", type=str, nargs='*')

    
    args = parser.parse_args()
    assert args.mode is not None
    if args.mode !="raw_data_read": assert args.outpfx is not None
    if args.infile is None:
        args.infile = "/afs/csail.mit.edu/u/r/rsingh/work/schema/data/slideseq/processed/puck_180430_1.h5ad"
    extra_args = dict([a.split("=") for a in args.extra]) if args.extra else {}

    
    if args.mode== "raw_data_read":
        read_raw_files_and_write_h5_files( args.outdir)


    if args.mode == "schema_kd_granule_cells":
        adata1 = SlideSeq.loadAnnData(args.infile)
        try:
            from schema import schema_qp
        except:
            sys.path.append(os.path.join(sys.path[0],'../../schema'))
            import schema_qp

        schema_qp.schema_loglevel = logging.WARNING
        
        fig, d3, schema_ret, min_corr1, min_corr2, scores, cell_loadings = processGranuleCellDensitySchema(adata1, extra_args)
        fig.tight_layout()
        fig.savefig("{0}_fig-KD.png".format(args.outpfx), dpi=300)
        fig.savefig("{0}_fig-KD.svg".format(args.outpfx))
        pickle.dump((d3[["xcoord","ycoord","kd","atlas_cluster"]], schema_ret, min_corr1, min_corr2, scores, cell_loadings),
                    open("{0}_func_output.pkl".format(args.outpfx), "wb"))


    if args.mode == "cca_kd_granule_cells":
        adata1 = SlideSeq.loadAnnData(args.infile)
        if "kd" not in adata1.obs.columns:
            adata1 = computeKernelDensityGranuleCells(adata1,
                                                      kd_fit_granule_only = int(extra_args.get("kd_fit_granule_only",1))==1,
                                                      kd_bw = float(extra_args.get("kd_bw",125)))
        from sklearn.cross_decomposition import CCA
        cca = CCA(1)
        cca.fit(adata1.X, adata1.obs["kd"])            
        cca_sgn = np.sign(scipy.stats.pearsonr(adata1.obs["kd"], cca.x_scores_[:,0])[0]) #flip signs if needed
        fig = generatePlotGranuleCellDensity(adata1.obs, cca_sgn*cca.x_scores_[:,0])
        fig.tight_layout()
        fig.savefig("{0}_fig-CCA.png".format(args.outpfx), dpi=300)
        fig.savefig("{0}_fig-CCA.svg".format(args.outpfx))
        pickle.dump((adata1.obs[["xcoord","ycoord","kd","atlas_cluster"]], cca.x_scores_[:,0], cca.x_loadings_[:,0], cca.y_scores_[:,0]),
                    open("{0}_CCA_output.pkl".format(args.outpfx), "wb"))



    if args.mode == "cca2step_kd_granule_cells":
        adata1 = SlideSeq.loadAnnData(args.infile)
        if "kd" not in adata1.obs.columns:
            adata1 = computeKernelDensityGranuleCells(adata1,
                                                      kd_fit_granule_only = int(extra_args.get("kd_fit_granule_only",1))==1,
                                                      kd_bw = float(extra_args.get("kd_bw",125)))

        #### adata1 = adata1[:,:40] ## FOR TESTING

        from sklearn.cross_decomposition import CCA
        cca1 = CCA(1)
        cca1.fit(adata1.X, adata1.obs["kd"])           
        cca1_sgn = np.sign(scipy.stats.pearsonr(adata1.obs["kd"],cca1.x_scores_[:,0])[0]) #flip signs if needed

        cca2 = CCA(1)
        cca2.fit(adata1.X, 1*(adata1.obs["atlas_cluster"]==1))            
        cca2_sgn = np.sign(scipy.stats.pearsonr(1*(adata1.obs["atlas_cluster"]==1),cca2.x_scores_[:,0])[0]) #flip signs if needed

        score1 = cca1_sgn*cca1.x_scores_[:,0]
        score2 = cca2_sgn*cca2.x_scores_[:,0]
        scorex = 0.5 * (score1/np.std(score1)  + score2/np.std(score2))

        scorex = scorex/np.sqrt(np.sum(scorex**2))

        loadings = np.matmul(np.transpose(adata1.X), scorex)
        intcpt = 0

        print("Flag 2320.01 ", scorex.shape, adata1.X.shape, loadings.shape, describe(scorex), describe(loadings))


        fig = generatePlotGranuleCellDensity(adata1.obs, scorex)
        fig.tight_layout()
        fig.savefig("{0}_fig-CCA2STEP.png".format(args.outpfx), dpi=300)
        fig.savefig("{0}_fig-CCA2STEP.svg".format(args.outpfx))
        pickle.dump((adata1.obs[["xcoord","ycoord","kd","atlas_cluster"]], scorex, loadings, intcpt),
                    open("{0}_CCA2STEP_output.pkl".format(args.outpfx), "wb"))



    if args.mode == "cca_schema_comparison_plot":
        cca_pkl_file = extra_args["cca_pkl_file"]
        schema_pkl_file = extra_args["schema_pkl_file"]
        cca_d3, cca_x_scores, _ , _  = pickle.load(open(cca_pkl_file,"rb"))
        schema_d3,  _, _, _, _, cell_loadings = pickle.load(open(schema_pkl_file,"rb"))
        #fig = doSchemaCCA_CellScorePlot(cca_d3, cca_x_scores, cell_loadings)
        fig = doSchemaCCA_CellScorePlot2(cca_d3, cca_x_scores, cell_loadings)
        fig.savefig("{0}_fig-Schema-CCA-cmp.png".format(args.outpfx), dpi=300)
        fig.savefig("{0}_fig-Schema-CCA-cmp.svg".format(args.outpfx))
        


    if args.mode == "generate_multipuck_gene_ranks":
        pkl_file_glob = extra_args["pkl_file_glob"]
        assert extra_args["data_type"].lower() in ["schema","cca"]
        data_type = extra_args["data_type"].lower()
        
        pkl_flist = glob.glob(pkl_file_glob)
        print("Flag 67.10 ", pkl_flist)
        assert len(pkl_flist) > 0
        
        L = []
        for f in pkl_flist:
            if data_type == "schema":
                _, _, _, _, scores, _ = pickle.load(open(f,"rb")) 
                L.append([a[0] for a in sorted(scores.items(), key=lambda v:v[1])]) #in schema rankings, low number means top-rank

            elif data_type == "cca":
                d3, cca_x_scores, cca_x_loadings, _ = pickle.load(open(f,"rb"))
                cca_sgn = np.sign(scipy.stats.pearsonr(d3["kd"],cca_x_scores)[0])

                puckid = f[f.index("180430"):][:8]
                adata1 = SlideSeq.loadAnnData("{0}/puck_{1}.h5ad".format(os.path.dirname(f), puckid))
                
                df = pd.DataFrame.from_dict({"gene": list(adata1.uns["Wo"].columns), "cca_scores": cca_sgn*cca_x_loadings})
                df = df.sort_values("cca_scores", ascending=False)
                L.append(list(df.gene.values))

        Nmax  = max(len(a) for a in L)
        print ("Flag 67.40 ", len(L), len(L[0]), Nmax)
        cons_score = {}
        active_set = set()
        for i in range(1,Nmax+1):
            currset = set.intersection(*[set(a[:i]) for a in L])
            if len(currset) > len(active_set):
                for s in currset-active_set:
                    cons_score[s] = len(currset)
                active_set = currset

        g = []; s = []
        for k,v in cons_score.items():
            g.append(k)
            s.append(v)

        pd.DataFrame.from_dict({"gene": g, "rank": s}).to_csv("{0}_generankings_dtype-{1}.csv".format(args.outpfx, data_type), index=False)
        



#!/usr/bin/env python

import pandas as pd
import numpy as np
import scipy, sklearn, os, sys, string, fileinput, glob, re, math, itertools, functools, copy, multiprocessing
import scipy.stats, sklearn.decomposition, sklearn.preprocessing, sklearn.covariance, sklearn.neighbors
from scipy.stats import describe
from scipy import sparse
import os.path
import scipy.sparse
from scipy.sparse import csr_matrix, csc_matrix
from sklearn.preprocessing import normalize
from collections import defaultdict


#### local directory imports ####
oldpath = copy.copy(sys.path)
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

from schema_base_config import *

sys.path = copy.copy(oldpath)
####


def get_louvain_clustering(mtx, num_neighbors=30):
        import igraph
        gNN = igraph.Graph()
        N = mtx.shape[0]
        gNN.add_vertices(N)
        schema_debug("Flag 192.30 ", mtx.shape)
        mNN = scipy.sparse.coo_matrix( sklearn.neighbors.kneighbors_graph(mtx, num_neighbors))
        schema_debug("Flag 192.40 ", mNN.shape)
        gNN.add_edges([ (i,j) for i,j in zip(mNN.row, mNN.col)])
        schema_debug("Flag 192.50 ")
        import leidenalg as la
        p1 = la.find_partition(gNN, la.ModularityVertexPartition)
        schema_debug("Flag 192.60 ")
        return   np.array([a for a in p1.membership])


    
class ScatterPlotAligner:
    """
    When seeded with a Mx2 matrix (that will be used for a scatter plot), 
        will allow you to map other Mx2 matrices so that the original scatter plot and the 
        new scatter plot will be in a similar orientation and positioning

    !!!!! DEPRECATED - do not use !!!!!
    """
    
    def __init__(self):
        """
        you will need to seed it before you can use it
        """
        self._is_seeded = False

        
    def seed(self, D):
        """
        the base matrix to be used for seeding. Should be a 2-d numpy type array. 
        IMPORTANT: a transformed version of this will be returned to you by the call. Use that for plots.

        Returns:  1) Kx2 transformed matrix and, 2) (xmin, xmax, ymin, ymax) to use with plt.xlim & plt.ylim
        """
        assert (not self._is_seeded) # Trying to seed an already-seeded ScatterPlotAligner
        assert D.shape[1] == 2

        self._is_seeded = True
        self._K = D.shape[0]
        
        mcd = sklearn.covariance.MinCovDet()
        mcd.fit(D)

        v = mcd.mahalanobis(D)
        self._valid_idx = (v < np.quantile(v,0.75)) #focus on the 75% of non-outlier data

        self._seed = D.copy()

        # translate and scale _seed so it's centered nicely around the _valid_idx's
        seed_mu = self._seed[self._valid_idx].mean(axis=0)        
        self._seed = self._seed - seed_mu
        
        pt_magnitudes = (np.sum((self._seed[self._valid_idx])**2, axis=1))**0.5
        
        d_std = pt_magnitudes.std()
        self._seed = self._seed * (2.0/d_std)  #have 1-SD pts be roughly at dist 2 from origin

        # get the bounding boxes
        u1 = 2*self._seed[self._valid_idx].min(axis=0)
        self._xmin, self._ymin = u1[0], u1[1]
        u2 = 2*self._seed[self._valid_idx].max(axis=0)
        self._xmax, self._ymax = u2[0], u2[1]

        return self._seed, (self._xmin, self._xmax, self._ymin, self._ymax)


    
    def map(self, D):
        """
        D is a 2-d numpy-type array to be mapped to the seed. require D.shape == seed.shape 

        IMPORTANT: seed() should be called first. 
        Subsequent calls to map() will transform their inputs to match the seed's orientation

        Returns:  a transformed version of D that you should use for plotting
        """
        assert self._is_seeded
        assert D.shape[0] == self._K
        assert D.shape[1] == 2
        
        dx = D.copy()

        # translate and scale dx so it's centered nicely around the _valid_idx's
        dx_mu = dx[self._valid_idx].mean(axis=0)
        dx = dx - dx_mu

        pt_magnitudes = (np.sum((dx[self._valid_idx])**2, axis=1))**0.5
        
        d_std = pt_magnitudes.std()
        dx = dx * (2.0/d_std)

        #get the rotation matrix
        # from https://igl.ethz.ch/projects/ARAP/svd_rot.pdf (pg5) and http://post.queensu.ca/~sdb2/PAPERS/PAMI-3DLS-1987.pdf

        H = np.matmul( dx[self._valid_idx].T, self._seed[self._valid_idx])
        #H = np.matmul( self._seed[self._valid_idx].T, dx[self._valid_idx])

        import numpy.linalg

        U,S,V =  numpy.linalg.svd(H, full_matrices=True)
        det_R0 = np.linalg.det(np.matmul(V, U.T))
        R1 = np.diagflat(np.ones(2))
        R1[1,1] = det_R0

        R  = np.matmul(V, np.matmul(R1, U.T))

        dx.iloc[:,:] = np.transpose( np.matmul(R, dx.T))

        return dx  #, (self._xmin, self._xmax, self._ymin, self._ymax)



def sparse_read_csv(filename, index_col=None, verbose=False):
    """
csr read of file. Does a 100-row load first to get data-types.
returns sparse matrix, rows and cols.
based on https://github.com/berenslab/rna-seq-tsne

#### Parameters

`filename`: `string`



`index_col`: `string` (default=`None`)

    name of column that serves as index_col. Passed unchanged to read_csv(...,index_col=index_col, ...)


`verbose: `boolean` (default=`False`)
    
    say stuff
    """
    
    small_chunk = pd.read_csv(filename, nrows=100)
    coltypes = dict(enumerate([a.name for a in small_chunk.dtypes.values]))

    indexL = []
    chunkL = []
    with open(filename) as file:
        for i,chunk in enumerate(pd.read_csv(filename, chunksize=3000, index_col=index_col, dtype=coltypes)):
            if verbose: print('.', end='', flush=True)
            if i==0:
                colsL = list(chunk.columns)
            indexL.extend(list(chunk.index))
            chunkL.append(sparse.csr_matrix(chunk.values.astype(float)))
            
        mat = sparse.vstack(chunkL, 'csr')
        if verbose: print(' done')
    return mat, np.array(indexL), colsL



def fast_csv_read(filename, *args, **kwargs):
    """
fast csv read. Like sparse_read_csv but returns a dense Pandas DF
    """
    
    small_chunk = pd.read_csv(filename, nrows=50)
    if small_chunk.index[0] == 0:
        coltypes = dict(enumerate([a.name for a in small_chunk.dtypes.values]))
        return pd.read_csv(filename, dtype=coltypes, *args, **kwargs)
    else:
        coltypes = dict((i+1,k) for i,k in enumerate([a.name for a in small_chunk.dtypes.values]))
        coltypes[0] = str
        return pd.read_csv(filename, index_col=0, dtype=coltypes, *args, **kwargs)


    

class SlideSeq:
    """
Utility class for Slide-Seq (Rodriques et al., Science 2019) data. Most methods are static.
    """
    
    @staticmethod
    def loadRawData(datadir, puckid, num_nmf_factors=100, prep_for_benchmarking=False):
        """
Load data for a particular puck, clean it up a bit and store as AnnData. For later use, also performs a NMF and stores those.
Borrows code from autoNMFreg_windows.py, provided with the Slide-Seq raw data.
        """
        from sklearn.preprocessing import StandardScaler
        
        puckdir = "{0}/Puck_{1}".format(datadir, puckid)
        beadmapdir = max(glob.glob("{0}/BeadMapping_*-*_????".format(puckdir)), key=os.path.getctime)
        schema_debug("Flag 314.001 ", beadmapdir)
        
        # gene exp
        gexp_file = "{0}/MappedDGEForR.csv".format(beadmapdir)
        dge = fast_csv_read(gexp_file, header = 0, index_col = 0)
        #  for faster testing runs, use below, it has just the first 500 cols of the gexp_file 
        ## dge = fast_csv_read("/tmp/a1_dge.csv", header = 0, index_col = 0)
        dge = dge.T
        dge = dge.reset_index()
        dge = dge.rename(columns={'index':'barcode'})
        schema_debug("Flag 314.010 ", dge.shape, dge.columns)
        
        # spatial location
        beadloc_file = "{0}/BeadLocationsForR.csv".format(beadmapdir)
        coords = fast_csv_read(beadloc_file, header = 0)
        coords = coords.rename(columns={'Barcodes':'barcode'})
        coords = coords.rename(columns={'barcodes':'barcode'})
        schema_debug("Flag 314.020 ", coords.shape, coords.columns)
        
        # Slide-Seq cluster assignments
        atlas_clusters_file = "{0}/AnalogizerClusterAssignments.csv".format(beadmapdir)
        clstrs = pd.read_csv(atlas_clusters_file, index_col=None)
        assert list(clstrs.columns) == ["Var1","x"]
        clstrs.columns = ["barcode","atlas_cluster"]
        clstrs = clstrs.set_index("barcode")
        schema_debug("Flag 314.030 ", clstrs.shape, clstrs.columns)
        
        df_merged = dge.merge(coords, right_on='barcode', left_on='barcode')
        df_merged = df_merged[ df_merged.barcode.isin(clstrs.index)]
        schema_debug("Flag 314.040 ", df_merged.shape, df_merged.columns)
            
        # remove sparse gene exp
        counts = df_merged.drop(['xcoord', 'ycoord'], axis=1)
        counts2 = counts.copy(deep=True)
        counts2 = counts2.set_index('barcode') #.drop('barcode',axis=1)
        counts2_okcols = counts2.sum(axis=0) > 0
        counts2 = counts2.loc[:, counts2_okcols]
        UMI_threshold = 5
        counts2_umis = counts2.sum(axis=1).values
        counts2 = counts2.loc[counts2_umis > UMI_threshold,:]
        schema_debug("Flag 314.0552 ", counts.shape, counts2.shape, counts2_umis.shape,isinstance(counts2, pd.DataFrame))
        
        #slide-seq authors normalize to have sum=1 across each bead, rather than 1e6
        cval = counts2_umis[counts2_umis>UMI_threshold]
        if not prep_for_benchmarking:
            counts2 = counts2.divide(cval, axis=0) #np.true_divide(counts2, counts2_umis[:,None])
            #counts2 = np.true_divide(counts2, counts2_umis[:,None])
                
            # this is also a little unusual, but I'm following their practice
            counts2.iloc[:,:] = StandardScaler(with_mean=False).fit_transform(counts2.values)
            schema_debug("Flag 314.0553 ", counts2.shape, counts2_umis.shape,isinstance(counts2, pd.DataFrame))
        
        coords2 = df_merged.loc[ df_merged.barcode.isin(counts2.index), ["barcode","xcoord","ycoord"]].copy(deep=True)
        coords2 = coords2.set_index('barcode') #.drop('barcode', axis=1)
        schema_debug("Flag 314.0555 ", coords2.shape,isinstance(coords2, pd.DataFrame))
        
        ok_barcodes = set(coords2.index) & set(counts2.index) & set(clstrs.index)
        schema_debug("Flag 314.060 ", coords2.shape, counts2.shape, clstrs.shape, len(ok_barcodes))
                
        if prep_for_benchmarking:
            return (counts2[counts2.index.isin(ok_barcodes)].sort_index(), coords2[coords2.index.isin(ok_barcodes)].sort_index(), clstrs[clstrs.index.isin(ok_barcodes)].sort_index())

        ## do NMF
        K1 = num_nmf_factors
        listK1 = ["P{}".format(i+1) for i in range(K1)]
        random_state = 17 #for repeatability, a fixed value
        model1 = sklearn.decomposition.NMF(n_components= K1, init='random', random_state = random_state, alpha = 0, l1_ratio = 0)
        Ho = model1.fit_transform(counts2.values)  #yes, slideseq code had Ho and Wo mixed up. Just following their lead here. 
        Wo = model1.components_

        schema_debug("Flag 314.070 ", Ho.shape, Wo.shape)
        
        Ho_norm = StandardScaler(with_mean=False).fit_transform(Ho)
        Ho_norm = pd.DataFrame(Ho_norm)
        Ho_norm.index = counts2.index
        Ho_norm.columns = listK1
        Wo = pd.DataFrame(Wo)
        Wo.index = listK1; Wo.index.name = "Factor"
        Wo.columns = list(counts2.columns)

        Ho_norm = Ho_norm[Ho_norm.index.isin(ok_barcodes)]
        Ho_norm = Ho_norm / Ho_norm.std(axis=0)

        schema_debug("Flag 314.080 ", Ho_norm.shape, Wo.shape)
        
        genexp = counts2[ counts2.index.isin(ok_barcodes)].sort_index()
        beadloc = coords2[ coords2.index.isin(ok_barcodes)].sort_index()
        clstrs = clstrs[ clstrs.index.isin(ok_barcodes)].sort_index()
        Ho_norm = Ho_norm.sort_index()

        schema_debug("Flag 314.090 ", genexp.shape, beadloc.shape, clstrs.shape, Ho_norm.shape, genexp.index[:5], beadloc.index[:5])

        beadloc["atlas_cluster"] = clstrs["atlas_cluster"]
        
        if "AnnData" not in dir():
            from anndata import AnnData

        adata = AnnData(X = genexp.values, obs = beadloc, uns = {"Ho": Ho_norm, "Ho.index": list(Ho_norm.index), "Ho.columns": list(Ho_norm.columns),
                                                                 "Wo": Wo, "Wo.index": list(Wo.index), "Wo.columns": list(Wo.columns)})
        return adata



    @staticmethod
    def loadAnnData(fpath):
        """
Import a h5ad file. Also deals with some scanpy weirdness when loading dataframes in the .uns slot
        """
        
        import matplotlib
        matplotlib.use("agg") #otherwise scanpy tries to use tkinter which has issues importing
        import scanpy as sc 

        adata = sc.read(fpath)
        for k in ["Ho","Wo"]:
            adata.uns[k] = pd.DataFrame(adata.uns[k], index=adata.uns[k + ".index"], columns= adata.uns[k + ".columns"])
        return adata

    
    
        
class SciCar:
    """
Utility class for Sci-Car (Cao et al., Science 2018) data. Most methods are static. 
    """

    
    @staticmethod
    def loadData(path, gsm_info, refdata):
        """
Load sci-CAR data format, as uploaded to GEO. Written for DOI:10.1126/science.aau0730 
but will probably work with other Shendure Lab datasets as well

#### Parameters

`path`: `string`

    directory where the files are


`gsm_info`:  `list` of (`string`, `string`, `string`, `function`, `function`) tuples

    Each tuple corresponds to one dataset (e.g. RNA-seq or ATAC-seq)
    The list should be in order of importance. Once a feature name for a modality shows up, 
    it will be ignored in subsequent tuples
    The cells will be taken as the intersection of "sample" column from the various cell files

    tuple = (name, modality, gsm_id, f_cell_filter, f_mdlty_filter)

       name: your name for the dataset. 
       modality: 'gene' or 'peak' for now
       gsm_id: GSMXXXXXX id
       f_cell_filter: None or a function that returns a boolean given a row-vector, used to filter
       f_mdlty_filter: None or a function that returns a boolean given a row-vector, used to filter


`refdata`: `string`: path to a tab-separated file

    Read as a dataframe containing Ensemble IDs ("ensembl_id"), TSS start/end etc. 

#### Returns

AnnData object where the cells are rows in X, the columns of dataset #1 are in .X, 
    the columns for remaining datasets are in .uns dicts keyed by given dataset name.
    The cell-line info is in .obs, while dataset #1's modality info is in .var
        """

        dref = pd.read_csv(refdata, sep="\t", low_memory=False)
        dref["tss_adj"] = np.where(dref["strand"]=="+", dref["txstart"], dref["txend"])
        dref = dref["ensembl_id,chr,strand,txstart,txend,tss_adj,map_location,Symbol".split(",")]
        
        if "AnnData" not in dir():
            from anndata import AnnData
        
        datasets = []
        cell_sets = []
        for nm, typestr, gsm, f_cell_filter, f_mdlty_filter in gsm_info:
            try:
                assert typestr in ['gene','peak']
                
                cells = pd.read_csv(glob.glob("{0}/{1}*_cell.txt".format(path, gsm))[0], low_memory=False)
                mdlty = pd.read_csv(glob.glob("{0}/{1}*_{2}.txt".format(path, gsm, typestr))[0], low_memory=False)

                
                cell_idx = np.full((cells.shape[0],),True)
                if f_cell_filter is not None:
                    cell_idx = (cells.apply(f_cell_filter, axis=1, result_type="reduce")).values
                    
                if typestr=="gene":
                    mdlty_idx = np.full((mdlty.shape[0],),True)
                    if f_mdlty_filter is not None:
                        mdlty_idx = (mdlty.apply(f_mdlty_filter, axis=1, result_type="reduce")).values

                    schema_debug ("Flag 324.112 filtered {0}:{1}:{2} mdlty".format(len(mdlty_idx), np.sum(mdlty_idx), mdlty_idx.shape))
                    
                    mdlty["gene_id"] = mdlty["gene_id"].apply(lambda v: v.split('.')[0])
                    mdlty_idx[~(mdlty["gene_id"].isin(dref["ensembl_id"]))] = False

                    schema_debug ("Flag 324.113 filtered {0}:{1}:{2} mdlty".format(len(mdlty_idx), np.sum(mdlty_idx), mdlty_idx.shape))
                    
                    mdlty["index"] = np.arange(mdlty.shape[0])
                    mdlty = (pd.merge(mdlty, dref, left_on="gene_id", right_on="ensembl_id", how="left"))
                    mdlty = mdlty.drop_duplicates("index").sort_values("index").reset_index(drop=True).drop(columns=["index"])

                    schema_debug ("Flag 324.114 filtered {0}:{1}:{2} mdlty".format(len(mdlty_idx), np.sum(mdlty_idx), mdlty_idx.shape))

                    def f_is_standard_chromosome(v):
                        try:
                            assert v[:3]=="chr" and ("_" not in v) and (v[3] in ['X','Y'] or int(v[3:])<=22 )
                            return True
                        except:
                            return False

                    mdlty_idx[~(mdlty["chr"].apply(f_is_standard_chromosome))] = False

                    schema_debug ("Flag 324.115 filtered {0}:{1}:{2} mdlty".format(len(mdlty_idx), np.sum(mdlty_idx), mdlty_idx.shape))
                    
                else:                    
                    mdlty_idx = np.full((mdlty.shape[0],),True)
                    if f_mdlty_filter is not None:
                        mdlty_idx = (mdlty.apply(f_mdlty_filter, axis=1, result_type="reduce")).values

                
                    
                # data = pd.DataFrame(data = scipy.io.mmread(glob.glob("{0}/{1}*_count.txt".format(path, gsm))[0]).T.tocsr().astype(np.float_),
                #                     index = cells['sample'],
                #                     columns = mdlty['mdlty_short_name'])

                data = scipy.io.mmread(glob.glob("{0}/{1}*_{2}_count.txt".format(path, gsm, typestr))[0]).T.tocsr().astype(np.float_)

                schema_debug ("Flag 324.12 read {0} cells, {1} mdlty, {2} data".format(cells.shape, mdlty.shape, data.shape))


                schema_debug ("Flag 324.13 filtered  {0}:{1}:{2} cells, {3}:{4}:{5} mdlty".format( len(cell_idx), np.sum(cell_idx), cell_idx.shape, len(mdlty_idx), np.sum(mdlty_idx), mdlty_idx.shape))
                    
                data = data[cell_idx, :]
                data = data[:, mdlty_idx]
                #data = data[cell_idx,:] # mdlty_idx]
                cells = cells[cell_idx].reset_index(drop=True)
                mdlty = mdlty[mdlty_idx].reset_index(drop=True)

                schema_debug ("Flag 324.14 filtered down to {0} cells, {1} mdlty, {2} data".format(cells.shape, mdlty.shape, data.shape))
                

                schema_debug ("Flag 324.15 filtered down to {0} cells, {1} mdlty, {2} data".format(cells.shape, mdlty.shape, data.shape))
                
                sortidx = np.argsort(cells['sample'].values)
                cells = cells.iloc[sortidx,:].reset_index(drop=True)
                data = data[sortidx, :]

                schema_debug ("Flag 324.17 \n {0} \n {1}".format( cells.head(2), data[:2,:2]))
                       
                datasets.append((nm, typestr, data, cells, mdlty))
                
                cell_sets.append(set(cells['sample'].values))
                
            except:
                raise 
                #raise ValueError('{0} could not be read in {1}'.format(nm, path))

        common_cells = list(set.intersection(*cell_sets))
        schema_debug ("Flag 324.20 got {0} common cells {1}".format( len(common_cells), common_cells[:10]))

        def logcpm(dx):
            libsizes = 1e-6 + np.sum(dx, axis=1)
            dxout = copy.deepcopy(dx)
            for i in range(dxout.shape[0]):
                i0,i1 = dxout.indptr[i], dxout.indptr[i+1]
                dxout.data[i0:i1] = np.log2(dxout.data[i0:i1]*1e6/libsizes[i] + 1)
                #for ind in range(i0,i1):
                #    dxout.data[ind] = np.log2( dxout.data[ind]*1e6/libsizes[i] + 1)
            return dxout

               
        for i, dx in enumerate(datasets):
            nm, typestr, data, cells, mdlty = dx
            
            cidx = np.in1d(cells["sample"].values, common_cells)
            schema_debug ("Flag 324.205 got {0} {1} {2} {3}".format( nm, cells.shape, len(cidx), np.sum(cidx)))
                    
            data = data[cidx,:]
            cells = cells.iloc[cidx,:].reset_index(drop=True)
            mdlty = mdlty.set_index('gene_short_name' if typestr=='gene' else 'peak')
            
            if i==0:
                cells = cells.set_index('sample')
                #adata = AnnData(X = logcpm(data), obs = cells.copy(deep=True), var = mdlty.copy(deep=True))
                adata = AnnData(X = data, obs = cells.copy(deep=True), var = mdlty.copy(deep=True))
                adata.uns["names"] = []

                schema_debug ("Flag 324.22 got X {0} obs {1} var {2} uns {3}".format( adata.X.shape, adata.obs.shape, adata.var.shape, list(adata.uns.keys())))
            else:
                for c in cells.columns:
                    if c in adata.obs.columns: continue
                    adata.obs[c] = cells[c]
                #adata.uns[nm + ".X"] = logcpm(data)
                adata.uns[nm + ".X"] = data
                adata.uns[nm + ".var"] = mdlty.copy(deep=True)
                adata.uns[nm + ".var.index"] = list(mdlty.index) #scanpy is annoying, it'll convert these to numpy matrices when writing
                adata.uns[nm + ".var.columns"] = list(mdlty.columns)

            adata.obs[typestr + "_log_sumcounts"] = np.log2(np.sum(data,axis=1)+1)
            adata.uns["names"].append(nm)
            adata.uns[nm + ".type"] = typestr
            adata.var_names_make_unique()
            schema_debug ("Flag 324.25 got X {0} obs {1} var {2} uns {3}".format( adata.X.shape, adata.obs.shape, adata.var.shape, list(adata.uns.keys())))
                   
        return adata

    @staticmethod
    def loadAnnData(fpath):
        import matplotlib
        matplotlib.use("agg") #otherwise scanpy tries to use tkinter which has issues importing
        import scanpy as sc 

        adata = sc.read(fpath)
        for k in adata.uns.keys():
            if k.endswith(".var") and isinstance(adata.uns[k], np.ndarray):
                schema_debug("Flag 343.100 hello", k, adata.uns.keys(), type(adata.uns[k]))
                adata.uns[k] = pd.DataFrame(adata.uns[k], index= adata.uns[k + ".index"], columns = adata.uns[k + ".columns"])
                schema_debug("Flag 343.102 ", k, adata.uns.keys(), type(adata.uns[k]))
                for c in  adata.uns[k].columns:
                    if c in ["id","start","end"]:
                        adata.uns[k][c] = adata.uns[k][c].astype(int)
                    if c in ["chr"]:
                        adata.uns[k][c] = adata.uns[k][c].astype(str)
        return adata

    @staticmethod
    def preprocessAnnData(adata, do_logcpm=True, valid_gene_minobs=0, valid_peak_minobs=0, valid_cell_mingenes=0):
        """
Preprocess sci-CAR data to remove too-sparse genes, peaks and cells. Also, convert to log(counts_per_million(..)) format

#### Parameters

`adata`: `AnnData`

    The dataframe containing Read as a dataframe containing Ensemble IDs ("ensembl_id"), TSS start/end etc. 


`do_logcpm`: `bool` 

    Convert peak and gene expression counts to log2 counts-per-million


`valid_gene_minobs`: `int`

    Only keep genes that show up in at least valid_gene_minobs cells


`valid_aux_minobs`: `int`

    Only keep peaks that show up in at least valid_peak_minobs cells


`valid_cell_mingenes`: `int`

    Only keep cells that have at least valid_cell_mingenes genes

#### Returns

copy of filtered anndata
         """

        valid_genes =  np.ravel((adata.X  > 0).sum(axis=0)) >= valid_gene_minobs
        adata = adata[:, valid_genes]
        
        valid_cells = np.ravel((adata.X > 0).sum(axis=1)) >= valid_cell_mingenes
        adata = adata[valid_cells, :]
        
        if "atac.X" in adata.uns:
            adata.uns["atac.X"] = adata.uns["atac.X"][ valid_cells, :]
            
            valid_peaks = np.ravel((adata.uns["atac.X"] > 0).sum(axis=0)) >= valid_peak_minobs
            adata.uns["atac.X"] = adata.uns["atac.X"][:, valid_peaks]
            adata.uns["atac.var"] = adata.uns["atac.var"][valid_peaks]
            adata.uns["atac.var.index"] = adata.uns["atac.var.index"][valid_peaks]
            
        adata2 = adata.copy()
         
        def logcpm(dx):
            libsizes = 1e-6 + np.sum(dx, axis=1)
            schema_debug ("Flag 3343.10 ",  libsizes.shape, libsizes.sum())
            dxout = dx #copy.deepcopy(dx)
            for i in range(dxout.shape[0]):
                i0,i1 = dxout.indptr[i], dxout.indptr[i+1]
                dxout.data[i0:i1] = np.log2(dxout.data[i0:i1]*1e6/libsizes[i] + 1)
                #for ind in range(i0,i1):
                #    dxout.data[ind] = np.log2( dxout.data[ind]*1e6/libsizes[i] + 1)
            return dxout
                 
        adata2.X = logcpm(adata2.X)
        if  "atac.X" in adata.uns:
            adata2.uns["atac.X"] = logcpm(adata2.uns["atac.X"])

        return adata2


    
    @staticmethod
    def getChrMapping(adata):    
        """
Get a mapping of genes/peaks to chromosomes and back

#### Parameters

`adata`: `AnnData` object

    output from loadData(...)

#### Returns
    gchr, chr2genes, pchr, chr2peaks : the first and third are integer vectors, the second and fourth are int->set(int) dicts. All gene and peak
      integers  refer to indexes 
        """
        chr2genes = defaultdict(set)
        chr2peaks = defaultdict(set)
        gchr = adata.var["chr"].astype(str).apply(lambda s: s.replace("chr",""))
        for i in range(adata.var.shape[0]):
            chr2genes[gchr[i]].add(i)
        pchr = adata.uns["atac.var"]["chr"].astype(str).apply(lambda s: s.replace("chr",""))
        for i in range(adata.uns["atac.var"].shape[0]):
            chr2peaks[pchr[i]].add(i)
        return gchr, chr2genes, pchr, chr2peaks


    @staticmethod
    def getPeakPosReGenes(gVar, peak, gene2chr):
        """
Given a peak, get its position vis-a-vis the genes in the genome

#### Parameters

`gVar`:  `pd.DataFrame`

    adata.var df from adata object


`peak`: `list` of size 3

   [chr, start, end] identifying the peak

`gene2chr`: `array of strings`

   for each gene idx, indicates which chromosome it is a part of 
 
#### Returns
    gVar.shape[0] x 5 nd-array, with 5 numbers describing the peak's posn re the gene (see code)
        """
        pchr, pstart, pend = peak
        schema_debug ("Flag 321.10012 ", peak)
        pchr = (str(pchr)).replace("chr","")


        #5 dims: 0: same chr, 1: pStart-gStart, 2: pEnd-gStart, 3: pStart-gEnd, 4: pEnd-gEnd
        pos = np.zeros((gVar.shape[0],5))
        pos[:,0] = np.where(gene2chr==pchr,1,0)
        pos[:,1] = np.where(gVar["strand"]=="+", gVar["txstart"]-pstart, pend-gVar["txend"]) #+ if pstart is upstream of txstart
        pos[:,2] = np.where(gVar["strand"]=="+", gVar["txstart"]-pend, pstart-gVar["txend"]) #+ if pend is upstream of txstart
        pos[:,3] = np.where(gVar["strand"]=="+", gVar["txend"]-pstart, pend-gVar["txstart"]) #+ if pstart is upstream of txend
        pos[:,4] = np.where(gVar["strand"]=="+", gVar["txend"]-pend, pstart-gVar["txstart"]) #+ if pend is upstream of txend

        assert np.sum((pos[:,2] > 0) & (pos[:,1] <= 0)) ==0
        assert np.sum((pos[:,1] > 0) & (pos[:,3] <= 0)) ==0
        assert np.sum((pos[:,3] < 0) & (pos[:,4] >= 0)) ==0
        assert np.sum((pos[:,4] < 0) & (pos[:,2] >= 0)) ==0

        # pos[:,1] = np.where(gVar["strand"]=="+", pstart-gVar["txstart"], gVar["txend"]-pend)
        # pos[:,2] = np.where(gVar["strand"]=="+", pend-gVar["txstart"], gVar["txend"]-pstart)
        # pos[:,3] = np.where(gVar["strand"]=="+", pstart-gVar["txend"], gVar["txstart"]-pend)
        # pos[:,4] = np.where(gVar["strand"]=="+", pend-gVar["txend"], gVar["txstart"]-pstart)
        return pos


    @staticmethod
    def fpeak_0_500(pos):
        """
        peak ends within 500bp upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>=0) & 
                         (pos[:,2] < 5e2)), 1, 0)

    @staticmethod
    def fpeak_500_2e3(pos):
        """
        peak ends within 500-2000bp upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>0) & 
                         (pos[:,2]> 5e2) & 
                         (pos[:,2] <= 2e3)), 1, 0)

    @staticmethod
    def fpeak_2e3_20e3(pos):
        """
        peak ends within 2k-20kb  upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>0) & 
                         (pos[:,2]> 2e3) & 
                         (pos[:,2] <= 20e3)), 1, 0)

    @staticmethod
    def fpeak_20e3_100e3(pos):
        """
        peak ends within 20-100kb upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>0) & 
                         (pos[:,2]> 20e3) & 
                         (pos[:,2] <= 100e3)), 1, 0)


    @staticmethod
    def fpeak_100e3_1e6(pos):
        """
        peak ends within 100kb-1Mb upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>0) & 
                         (pos[:,2]> 100e3) & 
                         (pos[:,2] <= 1e6)), 1, 0)

    @staticmethod
    def fpeak_1e6_10e6(pos):
        """
        peak ends within 1Mb-10Mb upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>0) & 
                         (pos[:,2]> 1e6) & 
                         (pos[:,2] <= 10e6)), 1, 0)


    @staticmethod
    def fpeak_10e6_20e6(pos):
        """
        peak ends within 10Mb-20Mb upstream of gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,2]>0) & 
                         (pos[:,2]> 10e6) & 
                         (pos[:,2] <= 20e6)), 1, 0)


    @staticmethod
    def fpeak_crossing_in(pos):
        """
        peak spans the TSS of the gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,1]>0) & 
                         (pos[:,2]<=0) &
                         (pos[:,4]>0)), 1, 0)


    @staticmethod
    def fpeak_inside(pos):
        """
        peak is between the start and end of the gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,1]<0) & 
                         (pos[:,2]<0) & 
                         (pos[:,3]>0) & 
                         (pos[:,4]>0)), 1, 0)


    @staticmethod
    def fpeak_crossing_out(pos):
        """
        peak is spands the txend of the gene
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,1]<0) & 
                         (pos[:,2]<0) & 
                         (pos[:,3]>0) & 
                         (pos[:,4]<0)), 1, 0)



    @staticmethod
    def fpeak_behind_1e3(pos):
        """
        peak starts within 1kb of gene end
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,3]<0) & 
                         (-pos[:,3] < 1e3)), 1, 0)


    @staticmethod
    def fpeak_behind_1e3_20e3(pos):
        """
        peak starts within 1-20kb of gene end
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,3]<0) & 
                         (-pos[:,3] > 1e3) &
                         (-pos[:,3] < 20e3)), 1, 0)


    @staticmethod
    def fpeak_behind_20e3_100e3(pos):
        """
        peak starts within 20kb-100kb of gene end
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,3]<0) & 
                         (-pos[:,3] > 20e3) &
                         (-pos[:,3] < 100e3)), 1, 0)

    @staticmethod
    def fpeak_behind_100e3_1e6(pos):
        """
        peak starts within 100kb-1Mb of gene end
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,3]<0) & 
                         (-pos[:,3] > 100e3) &
                         (-pos[:,3] < 1e6)), 1, 0)


    @staticmethod
    def fpeak_behind_1e6_10e6(pos):
        """
        peak starts within 1Mb-10Mb of gene end
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,3]<0) & 
                         (-pos[:,3] > 1e6) &
                         (-pos[:,3] < 10e6)), 1, 0)

    
    @staticmethod
    def fpeak_behind_10e6_20e6(pos):
        """
        peak starts within 10Mb-20Mb of gene end
        """
        return np.where(((pos[:,0]> 0) & 
                         (pos[:,3]<0) & 
                         (-pos[:,3] > 10e6) &
                         (-pos[:,3] < 20e6)), 1, 0)

    
    @staticmethod
    def fpeak_rbf_500(pos):
        """
        exp(-d/500)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/5e2)**2)), 0)


    @staticmethod
    def fpeak_rbf_1e3(pos):
        """
        exp(-d/1e3)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/1e3)**2)), 0)


    @staticmethod
    def fpeak_rbf_5e3(pos):
        """
        exp(-d/5e3)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/5e3)**2)), 0)


    @staticmethod
    def fpeak_rbf_20e3(pos):
        """
        exp(-d/20e3)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/20e3)**2)), 0)


    @staticmethod
    def fpeak_rbf_100e3(pos):
        """
        exp(-d/100e3)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/100e3)**2)), 0)


    @staticmethod
    def fpeak_rbf_1e6(pos):
        """
        exp(-d/1e6)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/1e6)**2)), 0)


    @staticmethod
    def fpeak_rbf_10e6(pos):
        """
        exp(-d/10e6)^2 if peak ends upstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,2]>=0), 
                                  np.exp(-((pos[:,2]/10e6)**2)), 0)


    @staticmethod
    def fpeak_behind_rbf_20e3(pos):
        """
        exp(-d/20e3)^2 if peak starts downstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,3]>=0), 
                                  np.exp(-((pos[:,3]/20e3)**2)), 0)


    @staticmethod
    def fpeak_behind_rbf_100e3(pos):
        """
        exp(-d/100e3)^2 if peak starts downstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,3]>=0), 
                                  np.exp(-((pos[:,3]/100e3)**2)), 0)


    @staticmethod
    def fpeak_behind_rbf_1e6(pos):
        """
        exp(-d/1e6)^2 if peak starts downstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,3]>=0), 
                                  np.exp(-((pos[:,3]/1e6)**2)), 0)


    @staticmethod
    def fpeak_behind_rbf_10e6(pos):
        """
        exp(-d/10e6)^2 if peak starts downstream of gene
        """
        return np.where((pos[:,0]> 0) & (pos[:,3]>=0), 
                                  np.exp(-((pos[:,3]/10e6)**2)), 0)
    


    fpeak_list_all = [ (fpeak_0_500.__func__, "fpeak_0_500"),
                       (fpeak_500_2e3.__func__, "fpeak_500_2e3"),
                       (fpeak_2e3_20e3.__func__, "fpeak_2e3_20e3"),
                       (fpeak_20e3_100e3.__func__, "fpeak_20e3_100e3"),
                       (fpeak_100e3_1e6.__func__, "fpeak_100e3_1e6"),
                       (fpeak_1e6_10e6.__func__, "fpeak_1e6_10e6"),
                       (fpeak_10e6_20e6.__func__, "fpeak_10e6_20e6"),
                       (fpeak_crossing_in.__func__, "fpeak_crossing_in"),
                       (fpeak_inside.__func__, "fpeak_inside"),
                       (fpeak_crossing_out.__func__, "fpeak_crossing_out"),
                       (fpeak_behind_1e3.__func__, "fpeak_behind_1e3"),
                       (fpeak_behind_1e3_20e3.__func__, "fpeak_behind_1e3_20e3"),
                       (fpeak_behind_20e3_100e3.__func__, "fpeak_behind_20e3_100e3"),
                       (fpeak_behind_100e3_1e6.__func__, "fpeak_behind_100e3_1e6"),
                       (fpeak_behind_1e6_10e6.__func__, "fpeak_behind_1e6_10e6"),
                       (fpeak_behind_10e6_20e6.__func__, "fpeak_behind_10e6_20e6"),
                       (fpeak_rbf_500.__func__, "fpeak_rbf_500"),
                       (fpeak_rbf_1e3.__func__, "fpeak_rbf_1e3"),
                       (fpeak_rbf_5e3.__func__, "fpeak_rbf_5e3"),
                       (fpeak_rbf_20e3.__func__, "fpeak_rbf_20e3"),
                       (fpeak_rbf_100e3.__func__, "fpeak_rbf_100e3"),
                       (fpeak_rbf_1e6.__func__, "fpeak_rbf_1e6"),
                       (fpeak_rbf_10e6.__func__, "fpeak_rbf_10e6"),
                       (fpeak_behind_rbf_20e3.__func__, "fpeak_behind_rbf_20e3"),
                       (fpeak_behind_rbf_100e3.__func__, "fpeak_behind_rbf_100e3"),
                       (fpeak_behind_rbf_1e6.__func__, "fpeak_behind_rbf_1e6"),
                       (fpeak_behind_rbf_10e6.__func__, "fpeak_behind_rbf_10e6"),
    ]
    


    @staticmethod
    def computeGeneByFpeakMatrix(adata, peak_func_list, chr_mapping = None, peakList = None, normalize_distwt=True, booleanPeakCounts = False):
        """
Compute a matrix that is nG x len(peak_func_list) with cell [i,j] being gene[i]'s dot-product with peak scores across all peaks, 
         subject to  peak wt as described by f_peak

#### Parameters

`adata` : AnnData object from loadData(...)

`peak_func_list` : `list` of  functions of the signature `pos: int`

`chr_mapping`:  `4-tuple` 

    Optional argument providing the output of `getChrMapping(..)`, for caching

`peakList`:  `list of int`

   Optional argument specifying which peak indexes to run over 

#### Returns
 
a   matrix of shape adata.var.shape[0] X len(peak_func_list)

        """

        if chr_mapping is None:
            gene2chr, chr2genes, peak2chr, chr2peaks = SciCar.getChrMapping(adata)
        else:
            gene2chr, chr2genes, peak2chr, chr2peaks = chr_mapping

        nCells, nGenes, nPeaks = adata.shape[0], adata.shape[1], adata.uns["atac.X"].shape[1]
        try:
            gXt = adata.X.T.tocsr()
        except:
            gXt = adata.X.T

        if booleanPeakCounts and gXt.shape[0] > 0:
            gXt = gXt > np.median(gXt, axis=0) ## UNDO ????

        pXt = adata.uns["atac.X"].T.tocsr()
        gVar = adata.var[["chr","strand","txstart","txend","tss_adj"]]
        pVar = adata.uns["atac.var"][["chr","start","end"]]

        k = len(peak_func_list)
        
        g2p = np.zeros((nGenes,k))
        g2p_wts = np.zeros((nGenes,k))

        if peakList is not None:
            plist = peakList
        else:
            plist = range(nPeaks)

        for i in plist:
            v = SciCar.getPeakPosReGenes(gVar, pVar.values[i,:], gene2chr)

            pXti =  np.ravel(pXt[i,:].todense())

            pXti_positive = (pXti > 0).astype(int)
            pXti_ones = np.ones_like(pXti)

            schema_debug ("Flag 2.0008 ",i, flush=True)

            for j,f in enumerate(peak_func_list):
                distwt = f(v)
                #schema_debug("Flag 2.0010 ", i,j,np.std(distwt), np.mean(distwt)) 
                if np.sum(distwt)>1e-12:
                    if normalize_distwt: distwt = distwt/np.mean(distwt)
                    G = gXt.copy()
                    try:
                        G.data *= distwt.repeat(np.diff(G.indptr))
                    except:
                        G = G*distwt[:,None]
                    # schema_debug ("Flag 2.0201 ", len(distwt), np.sum(G.data), G.shape, 
                    #        np.sum(gXt.data),
                    #        G.data.shape, len(np.diff(G.indptr)),
                    #       pXt.shape)
                    gw = G.dot(pXti).ravel()
                    if booleanPeakCounts:
                        gw = G.dot(pXti_positive).ravel()
                    g2p[:,j] += gw/nCells

                    gw_ones = G.dot(pXti_ones).ravel()
                    g2p_wts[:,j] += gw_ones/nCells

                    # schema_debug ("Flag 2.0320 ", G.shape, distwt.shape, gXt.shape, pXt.shape, gw.shape, g2p.shape)
                
        return (g2p, g2p_wts)


    


    @staticmethod
    def aggregatePeaksByGenes(adata, peak_func, chr_mapping = None):
        """
Compute a matrix that is nCells x nGenes with cell [i,j] being expression of peaks around gene[i] in cell [i], 
         subject to  peak wt as described by f_peak

#### Parameters

`adata` : AnnData object from loadData(...)

`peak_func` : function of the signature `pos: int`

`chr_mapping`:  `4-tuple` 

    Optional argument providing the output of `getChrMapping(..)`, for caching

#### Returns
 
a   matrix of shape adata.shape

        """
        
        
        if chr_mapping is None:
            gene2chr, chr2genes, peak2chr, chr2peaks = SciCar.getChrMapping(adata)
        else:
            gene2chr, chr2genes, peak2chr, chr2peaks = chr_mapping

        nCells, nGenes, nPeaks = adata.shape[0], adata.shape[1], adata.uns["atac.X"].shape[1]

        pX = adata.uns["atac.X"]

        exp_pX = np.exp(pX)  #data was log1p'd before. We'll add the cpm counts and then re-log1p it 
        gVar = adata.var[["chr","strand","txstart","txend","tss_adj"]]
        pVar = adata.uns["atac.var"][["chr","start","end"]]

        c2g = np.zeros(nCells, nGenes)
        c2g_tmp = np.zeros(nCells, nGenes)
        
        plist = range(nPeaks)

        for i in plist:
            c2g_tmp[:] = 0
            
            v = SciCar.getPeakPosReGenes(gVar, pVar.values[i,:], gene2chr)
            c2g_tmp += peak_func(v)
            p_i =  np.exp(np.ravel(pX[i,:].todense()))
            c2g_tmp *= p_i[:,None]

            c2g += c2g_tmp

        c2g = np.log1p(c2g)
        return c2g

    





def cmpSpearmanVsPearson(d1, type1="numeric", d2=None, type2=None, nPointPairs=2000000, nRuns=1):

    N, K1 = d1.shape[0], d1.shape[1]
    K2 = 0
    if d2 is not None:
        assert d2.shape[0] == N
        K2 = d2.shape[1]
        
    if nPointPairs is None: assert nRuns==1

    schema_debug ("Flag 676.10 ", N, K1, K2, nPointPairs, type1, type2)
    
    corrL = []

    for nr in range(nRuns):
        if nPointPairs is not None:
            j_u = np.random.randint(0, N, int(3*nPointPairs))
            j_v = np.random.randint(0, N, int(3*nPointPairs))
            valid = j_u < j_v  #get rid of potential duplicates (x1,x2) and (x2,x1) as well as (x1,x1)
            i_u = (j_u[valid])[:nPointPairs]
            i_v = (j_v[valid])[:nPointPairs]
        else:
            x = pd.DataFrame.from_records(list(itertools.combinations(range(N),2)), columns=["u","v"])
            x = x[x.u < x.v]
            i_u = x.u.values
            i_v = x.v.values

        schema_debug ("Flag 676.30 ", i_u.shape, i_v.shape, nr)

        dL = []
        for g_val, g_type in [(d1, type1), (d2, type2)]:
            if g_val is None:
                dL.append(None)
                continue

            dg = []
            for ii in np.split(np.arange(i_u.shape[0]), int(i_u.shape[0]/2000)):
                ii_u = i_u[ii]
                ii_v = i_v[ii]

                if g_type == "categorical":
                    dgx = 1.0*( g_val[ii_u] != g_val[ii_v]) #1.0*( g_val[i_u].toarray() != g_val[i_v].toarray())
                elif g_type == "feature_vector":
                    schema_debug (g_val[ii_u].shape, g_val[ii_v].shape)
                    dgx = np.ravel(np.sum(np.power(g_val[ii_u].astype(np.float64) - g_val[ii_v].astype(np.float64),2), axis=1))
                else:  #numeric
                    dgx = (g_val[ii_u].astype(np.float64) - g_val[ii_v].astype(np.float64))**2   #(g_val[i_u].toarray() - g_val[i_v].toarray
                dg.extend(dgx)
            schema_debug ("Flag 676.50 ", g_type, len(dg))
            dL.append(np.array(dg))

        schema_debug ("Flag 676.60 ")
        dg1, dg2 = dL[0], dL[1]
        
        if d2 is None:
            rp = scipy.stats.pearsonr(dg1, scipy.stats.rankdata(dg1))[0]
            rs = None
        else:
            rp = scipy.stats.pearsonr(dg1, dg2)[0]
            rs = scipy.stats.spearmanr(dg1, dg2)[0]

        schema_debug ("Flag 676.80 ", rp, rs)
        corrL.append( (rp,rs))

    return corrL











# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os.path
import sys
from datetime import datetime

import matplotlib
matplotlib.use('agg')

#sys.path[:0] = [os.path.dirname(os.path.abspath(__file__)) + "/../../schema/"]
#import schema_qp

sys.path[:0] = [os.path.dirname(os.path.abspath(__file__)) + "/../../"]
import schema


# -- Project information -----------------------------------------------------

project = 'Schema'
copyright = '2020, Rohit Singh, Ashwin Narayan, Brian Hie'
author = 'Rohit Singh, Ashwin Narayan, Brian Hie'

# The full version, including alpha/beta/rc tags
release = '0.0.5'

master_doc = 'index'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [ 'sphinx.ext.todo', 'sphinx.ext.viewcode', 'sphinx.ext.autodoc']


# Add any paths that contain templates here, relative to this directory.
templates_path = ['../_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

html_theme_options = {
    'logo_only':  True,
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['../_static']
html_logo = '../_static/Schema-webpage-logo-2-blue.png'

autoclass_content = 'both'


# def skip(app, what, name, obj, would_skip, options):
#     if name == "__init__":
#         return False
#     return would_skip

# def setup(app):
#     app.connect("autodoc-skip-member", skip)



#!/usr/bin/env python


###################################################################
## Primary Author:  Rohit Singh rsingh@alum.mit.edu
## Co-Authors: Ashwin Narayan, Brian Hie {ashwinn,brianhie}@mit.edu
## License: MIT
## Repository:  http://github.io/rs239/schema
###################################################################

import pandas as pd
import numpy as np
import scipy, sklearn
import os, sys, string, fileinput, glob, re, math, itertools, functools, copy, logging
import sklearn.decomposition, sklearn.preprocessing, sklearn.linear_model, sklearn.covariance
import cvxopt

#### local directory imports ####
oldpath = copy.copy(sys.path)
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

from schema_base_config import *

sys.path = copy.copy(oldpath)
####


class SchemaQP:
    """Schema is a tool for integrating simultaneously-assayed data modalities

    The SchemaQP class provides a sklearn type fit+transform API for constrained
    affine transformations of input datasets such that the transformed data is
    in agreement with all the input datasets.
"""

    
    def __init__(self, min_desired_corr=0.99, mode="affine", params={}):
        """

        :param min_desired_corr: 
            This parameter controls the severity of the primary modality's
            transformation, specifying the minimum required correlation
            between distances in the original space and those in the
            transformed space. It thus controls the trade-off between
            deviating further away from the primary modality's original
            representation and achieving greater agreement with the
            secondary modalities. Values close to one result in lower
            distortion of the primary modality while those close to zero
            enable transformations offering greater agreement
            between the modalities.

            RECOMMENDED VALUES: In typical single-cell use cases, high
            values (> 0.80) will probably work best. With these, the
            distortion will be low, but still be enough for Schema to
            extract relevant information from the secondary modalities.
            Furthermore, the feature weights computed by Schema should
            still be quite infromative.

            The default value of 0.99 is a safe choice to start with; it
            poses low risk of deviating too far from the primary modality.

            Later, you can experiment with a range of values (e.g., 0.95
            0.90, 0.80), or use feature-weights aggregated across an
            ensemble of choices. Alternatively, you can use
            cross-validation to identify the best setting

        :type min_desired_corr: float in [0,1)

        :param mode: 
            Whether to perform a general affine transformation or just a
            scaling transformation

            * `affine` first does a mapping to PCA or NMF space (you can
              specify num_top_components via the `params` argument). Schema does
              a scaling transform in the mapped space and then converts everything
              back to the regular space. The final result is thus an affine
              transformation in the regular space.

            * `scale` does not do a PCA or NMF mapping, and directly applies
              the scaling transformation. **Note**: This can be slow if
              the primary modality's dimensionality is over 100.


            RECOMMENDED VALUES: `affine` is the default. You may need `scale` 
            only in certain cases:

            * You have a limited number of features on which you directly
              want Schema to compute feature-weights.

            * You want to do a change-of-basis transform other PCA or NMF.
              If so, you will need to do that yourself and then call
              SchemaQP with the transformed primary dataset with
              mode='scale'.


        :type mode: string

        :param params: 
             Dictionary of key-value pairs, specifying additional
             configuration parameters. Here are the important ones:

               * `decomposition_model`: "pca" or "nmf" (default=pca)

               * `num_top_components`: (default=50) number of PCA (or NMF)
                 components to use when mode=="affine". We recommend this
                 setting be <= 100. Schema's runtime is quadratic in this
                 number.

             You can ignore the rest on your first pass; the default
             values are pretty reasonable:

               * `dist_npairs`: (default=2000000). How many pt-pairs to use
                 for computing pairwise distances. value=None means
                 compute exhaustively over all n*(n-1)/2 pt-pairs. Not
                 recommended for n>5000.  Otherwise, the given number of
                 pt-pairs is sampled randomly. The sampling is done in a
                 way in which each point will be represented roughly
                 equally.

               * `scale_mode_uses_standard_scaler`: 1 or 0 (default=0),
                 apply the standard scaler in the scaling mode

               * `do_whiten`: 1 or 0 (default=1). When mode=="affine",
                 should the change-of-basis loadings be made 1-variance?

        :type params: dict

        :returns: A SchemaQP object on which you can call fit(...), transform(...) or fit_transform(....).
"""
        if not (mode in ['scale', 'affine']): raise ValueError("'mode' must be one of ['affine','scale']")
        if not (0 <= min_desired_corr < 1): raise ValueError("'min_desired_corr' must be in the range [0,1)") 

        self._mode = mode
        self._w_max_to_avg = 1000 
        schema_info ("Flag 456.10  ", self._w_max_to_avg)
        self._min_desired_corr = min_desired_corr
        self._std_scaler = None
        self._decomp_mdl = None
        self._orig_decomp_mdl = None

        # defaults
        self._params = {"decomposition_model": "nmf",
                        "num_top_components": 20, 
                        "dist_npairs": 2000000,
                        "d0_dist_transform": None,
                        "secondary_data_dist_transform_list": None,
                        "d0_orig_transformed_corr": None,
                        "dist_df_frac": 1.0,
                        "scale_mode_uses_standard_scaler": 0,
                        "do_whiten": 1,
                        "require_nonzero_lambda": 0,
                        "d0_type_is_feature_vector_categorical": 0,
        }


        self._params.update(params) # overwrite with user settings
        
        if not (self._params["decomposition_model"].lower() in ["pca","nmf"]):
            raise ValueError("For change-of-basis transforms other than PCA/NMF, please do the transformation yourself and use SchemaQP with mode='scale'")
        
        if not (self._params['num_top_components'] is None or self._params["num_top_components"] >= 2): raise ValueError('Need num_top_components >= 2')



    def reset_mincorr_param(self, min_desired_corr):
        """Reset the min_desired_corr. 

        Useful when you want to iterate over multiple choices of this parameter
        but want to re-use the computed PCA or NMF change-of-basis transform.

        :param min_desired_corr: 
            The new value of minimum required correlation between original and transformed distances

        :type min_desired_corr: float in [0,1)

"""
        if min_desired_corr is None or not(0 <= min_desired_corr < 1): raise ValueError("'min_desired_corr' must be between 0 and 1")
        self._min_desired_corr = min_desired_corr



    def reset_maxwt_param(self, w_max_to_avg):
        """ Reset the w_max_to_avg param

        :type w_max_to_avg: float

        :param w_max_to_avg: 
            The upper-bound on the ratio of Schema weights (w's) largest
            element to w's avg element.  Making it large will allow This
            parameter controls the 'deviation' in feature weights and make it
            large will allow for more severe transformations.

            **Handle with care:** We recommend keeping this parameter at its
            default value (1000); that keeps this constraint very loose and
            ensures that min_desired_corr remains the binding constraint.
            Later, as you get a better sense for the right min_desired_corr
            values for your data, you can experiment with this too.  To really
            constrain this, set it in the (1-5] range, depending on how many
            features you have.

"""
        if w_max_to_avg is None or w_max_to_avg <= 1: raise ValueError("'w_max_to_avg' must be either None or greater than 1")
        self._w_max_to_avg = w_max_to_avg
        
    
        
    
    def fit(self, d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list = None, secondary_data_dist_kernels=None, d0 = None, d0_dist_transform=None):
        """Compute the optimal Schema transformation, first performing a
        change-of-basis transformation if required.

        Given the primary dataset `d` and a list of secondary datasets, fit a
        linear transformation (`d_new`) such that the correlation between
        squared pairwise distances in `d_new` and those in secondary datasets is
        maximized while the correlation between the original `d` and the transformed `d_new`
        remains above min_desired_corr.

        The first three arguments are required, the next is useful, and
        the rest should be rarely used.

        :type d: Numpy 2-d `array` or Pandas `dataframe`

        :param d: 
            The primary dataset (e.g. scanpy/anndata's .X).

            The rows are observations (e.g., cells) and the cols are variables (e.g.,
            gene expression).  The default distance measure computed is L2:
            sum((point1-point2)**2). Also see `d0_dist_transform`.


        :type secondary_data_val_list: list of 1-d or 2-d Numpy arrays or Pandas series, each with same number of rows as `d`

        :param secondary_data_val_list: 
            The secondary datasets you want to align the primary data towards.  

            Columns in Anndata .obs or .obsm variables work well.


        :type secondary_data_type_list: list of strings

        :param secondary_data_type_list: 
            The datatypes of the secondary modalities.

            Each element of the list can be one of `numeric,
            feature_vector, categorical, feature_vector_categorical`.  
            The list's length should match the length
            of secondary_data_val_list


            * `numeric`: one floating-pt value for each observation.  The
              default distance measure is Euclidean: (point1-point2)^2
 
            * `feature_vector`: a k-dimensional vector for each
              observation.  The default distance measure is Euclidean:
              sum_{i}((point1[i]-point2[i])^2)

            * `categorical`: a label for each observation.  The default
              distance measure checks for equality: 1*(val1!=val2)

            * `feature_vector_categorical`: a vector of labels for each
              observation.  Each column can take on categorical values, so
              the distance between two points is
              sum_{i}(point1[i]==point2[i])


        :type secondary_data_wt_list: list of floats, **optional**

        :param secondary_data_wt_list: 
            User-specified wts for each secondary dataset (default= list of 1's)

            If specified, the list's length should match the length of
            secondary_data_val_list. When multiple secondary modalities are
            specified, this parameter allows you to control their relative
            weight in seeking an agreement with the primary.

            **Note**: you can try to get a mapping that *disagrees* with a
            dataset_info instead of *agreeing*.  To do so, pass in a
            negative number (e.g., -1) here. This works even if you have
            just one secondary dataset

        :type secondary_data_dist_kernels: list of functions, **optional**

        :param secondary_data_dist_kernels: 
            The transformations to apply on
            secondary dataset's L2 distances before using them for correlations.

            If specified, the length of the list should match that of
            `secondary_data_val_list`. 
            Each function should take a non-negative float and return a non-negative float. 

            **Handle with care:** Most likely, you don't need this parameter.

        :type d0: A 1-d or 2-d Numpy array, **optional** 

        :param d0: 
            An alternative representation of the primary dataset.

            This is useful if you want to provide the primary dataset in two
            forms: one for transforming and another one for computing pairwise
            distances to use in the QP constraint; if so, `d` is used for the
            former, while `d0` is used for the latter.
            If specified, it should have the same number of rows as `d`.

            **Handle with care:** Most likely, you don't need this parameter.


        :type d0_dist_transform: float -> float function, **optional**

        :param d0_dist_transform: 
            The transformation to apply on d or d0's L2
            distances before using them for correlations.

            This function should take a non-negative float as input and return a non-negative float.

            **Handle with care:** Most likely, you don't need this parameter.

        :returns: None
         """
        
        if not (d.ndim==2): raise ValueError('d should be a 2-d array')
        if not np.isfinite(d).all(): raise ValueError('d should not have any NaN or inf values')
        
        if not (len(secondary_data_val_list) >0): raise ValueError('secondary_data_val_list can not be empty')

        if not (len(secondary_data_val_list)==len(secondary_data_type_list)):
            raise ValueError('secondary_data_type_list should have the same length as secondary_data_val_list')
        
        if not (secondary_data_wt_list is None or len(secondary_data_wt_list)==len(secondary_data_val_list)):
            raise ValueError('secondary_data_wt_list should have the same length as secondary_data_val_list')

        for i in range(len(secondary_data_val_list)):
            if not (secondary_data_type_list[i] in ['categorical','numeric','feature_vector', 'feature_vector_categorical']):
                raise ValueError('{0}-th entry in secondary_data_type_list is invalid'.format(i+1))
             
            if not (secondary_data_val_list[i].shape[0] == d.shape[0]):
                raise ValueError('{0}-th entry in secondary_data_val_list has incorrect rows'.format(i+1))
            
            if not ((secondary_data_type_list[i]=='categorical' and secondary_data_val_list[i].ndim==1) or
                    (secondary_data_type_list[i]=='numeric' and secondary_data_val_list[i].ndim==1) or
                    (secondary_data_type_list[i]=='feature_vector' and secondary_data_val_list[i].ndim==2) or
                    (secondary_data_type_list[i]=='feature_vector_categorical' and secondary_data_val_list[i].ndim==2)):
                raise ValueError('{0}-th entry in secondary_data_val_list does not match specified type'.format(i+1))

            
        if not (d0 is None or d0.shape[0] == d.shape[0]): raise ValueError('d0 has incorrect rows')

            

        self._params["d0_dist_transform"] = d0_dist_transform
        self._params["secondary_data_dist_transform_list"] = secondary_data_dist_kernels
        
        if self._mode=="scale":
            self._fit_scale(d, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
        else: #affine
            self._fit_affine(d, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)



            
            
    def transform(self, d):
        """Given a dataset `d`, apply the fitted transform to it

        :type d: Numpy 2-d array

        :param d: 
            The primary modality data on which to apply the transformation. 

            `d` must have with same number of columns as in `fit(...)`.
            The rows are observations (e.g., cells) and the cols are variables (e.g., gene expression).


        :returns: a 2-d Numpy array with the same shape as `d`
 """
        
        if self._mode=="scale":
            if not (d.shape[1] == len(self._wts)): raise ValueError('Number of columns in d is incorrect')
            dx = self._std_scaler.transform(d)
            return np.multiply(dx, np.sqrt(self._wts))
        
        else: #affine
            if not (d.shape[1] == self._decomp_mdl.components_.shape[1]): raise ValueError('Number of columns in d is incorrect')
            dx = self._decomp_mdl.transform(d)
            return np.multiply(dx, np.sqrt(self._wts))

        

        
    def fit_transform(self, d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list = None, secondary_data_dist_kernels = None, d0 = None, d0_dist_transform=None):
        """
        Calls fit(..) with exactly the arguments given; then calls transform(d).
        See documentation for fit(....) and transform(...) respectively.

        """

        self.fit(d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list, secondary_data_dist_kernels, d0, d0_dist_transform)
        return self.transform(d)



    def feature_weights(self):
        """
        Return the feature weights computed by Schema

        If SchemaQP was initialized with `mode=scale`, the weights
        returned are directly the weights from the quadratic programming
        (QP), with a weight > 1 indicating the feature was up-weighted.


        However, if `mode=affine` was used, the QP-computed weights
        correspond to columns of the PCA or NMF decomposition. In that
        case, this functions maps them back to the primary dataset's
        features. This is done by a 'softmax'-type summation of loadings
        across the PCA/NMF columns, with each column's weight proportional
        to exp(QP wt). 

        If you'd like to build your own mapping from
        PCA/NMF weights to gene weights, look at the code for this
        function. In particular, you'll need the QP weights (in `self._wts`( and the
        PCA/NMF model (in `self._decomp_mdl`)

        *returns* : a vector of floats, the same size as the primary dataset's dimensionality
"""
        if self._mode == "scale":
            return self._wts
        else:
            w = np.exp(self._wts)  # exponentiation of wts so the highest-wt features really stand out
            w[ self._wts <= 1] = 0 # ignore features with wt <=1 (these were not up-weighted)
            w = w/np.sum(w) # normalize

            if self.params['decomposition_model'] == 'nmf':
                feat_wts = sum([ w[i]*self._decomp_mdl.components_[i,:] for i in range(len(w))])
                
            else: #in PCA, loadings are unique upto a sign, so when adding them up, we just take the magnitudes 
                feat_wts = sum([ w[i]*np.abs(self._decomp_mdl.components_[i,:]) for i in range(len(w))])

            return feat_wts

        

    ###################################################################
    ######## "private" methods below. Not that Python cares... ########
    ###################################################################
    def _getDistances(self, D, d0_in, G, nPointPairs):
        """
        Compute the various distances between point pairs in D
        D is a NxK Numpy type 2-d array, with N points, each of K dimensions
        d0_in could be None, in which D is used to compute distances for the original space.
           Otherwise, d0_in is Nxr (r>=1) array and original-space distances are computed from this. 
        G is list, with each entry a 3-tuple: (g_val, g_type, gamma_g)
           each 3-tuple corresponds to one dimension of side-information
             g_val is Nx1 Numpy type 1-d vector of values for the N points, in the same order as D
             g_type is "numeric", "categorical", or "feature_vector"
             gamma_g is the relative wt you want to give this column. You can leave 
               it as None for all 3-tuples, but not for just some. If you leave them
               all as None, the system will set wts to 1  
        nPointPairs is the number of point pairs you want to evaluate over. If it is None,
             the system will generate over all possible pairs
        """
        ##############################
        # symbology
        # uv_dist = nPointPairs x K matrix of squared distances between point pairs, along K dimensions
        # uv_dist_mean = 1 x K vector: avg of uv_dist along each dimension 
        # uv_dist_centered = nPointPairs x K matrix with uv_dist_mean subtracted from each row
        # d0 = nPointPairs x 1 vector: squared distances in the current space
        # z_d0 = z-scored version of  d0
        # dg = nPointPairs x 1 vector: squared distances (or class-match scores) for grouping g
        # z_dg = z-scored version of dg
        ##############################
        
        schema_info ("Flag 232.12 ", len(G), G[0][0].shape, G[0][1], G[0][2])
        gamma_gs_are_None = True
        for i,g in enumerate(G):
            g_val, g_type, gamma_g = g
                
            assert g_type in ['categorical','numeric','feature_vector']
            
            if i==0 and gamma_g is not None:
                gamma_gs_are_None = False
            if gamma_g is None: assert gamma_gs_are_None
            if gamma_g is not None: assert (not gamma_gs_are_None)
            

        N, K = D.shape[0], D.shape[1]

        if nPointPairs is not None:
            j_u = np.random.randint(0, N, int(3*nPointPairs))
            j_v = np.random.randint(0, N, int(3*nPointPairs))
            valid = j_u < j_v  #get rid of potential duplicates (x1,x2) and (x2,x1) as well as (x1,x1)
            i_u = (j_u[valid])[:nPointPairs]
            i_v = (j_v[valid])[:nPointPairs]
        else:
            x = pd.DataFrame.from_records(list(itertools.combinations(range(N),2)), columns=["u","v"])
            x = x[x.u < x.v]
            i_u = x.u.values
            i_v = x.v.values
            
        # NxK matrix of square distances along dimensions
        if int(self._params.get("d0_type_is_feature_vector_categorical",1)) > 0.5:
            uv_dist = (D[i_u,:] == D[i_v,:]).astype(np.float64)
        else:
            uv_dist = (D[i_u,:].astype(np.float64) - D[i_v,:].astype(np.float64))**2 


        # scale things so QP gets ~1-3 digit numbers, very large or very small numbers cause numerical issues         
        uv_dist = uv_dist/ (np.sqrt(uv_dist.shape[0])*uv_dist.ravel().mean())  
        
        # center uv_dist along each dimension
        uv_dist_mean = uv_dist.mean(axis=0) #Kx1 vector, this is the mean of (u_i - v_i)^2, not sqrt((u_i - v_i)^2) 
        uv_dist_centered = uv_dist - uv_dist_mean
        
        # square distances in the current metric space
        if d0_in is None:
            d0 = uv_dist.sum(axis=1) # Nx1
        else:
            dx0 = (d0_in[i_u] - d0_in[i_v])**2
            if dx0.ndim==2:
                d0 = dx0.sum(axis=1)
            else:
                d0 = dx0

        d0_orig = d0.copy()            
        f_dist_d0 = self._params["d0_dist_transform"]
        self._params["d0_orig_transformed_corr"] = 1.0

        if f_dist_d0 is not None:
            d0 = f_dist_d0(d0)
            cr = (np.corrcoef(d0_orig, d0))[0,1]
            if cr<0:
                # it's an error if you specify a d0 such that corr( pairwisedist(d), pairwise(d0)) is negative to begin with
                schema_error("d0_dist_transform inverts the correlation structure {0}.".format(cr))
                raise ValueError("""d0_dist_transform inverts the correlation structure. Aborting...
                                    It's an error if you specify a d0 such that corr( pairwisedist(d), pairwise(d0)) is negative to begin with""")
            self._params["d0_orig_transformed_corr"] = cr

            
        z_d0  = (d0 - d0.mean())/d0.std()

        schema_info("Flag 2090.20 Initial corr to d0", np.corrcoef(uv_dist_centered.sum(axis=1), z_d0))
        
        
        l_z_dg = []
        for ii, gx in enumerate(G):
            g_val, g_type, gamma_g = gx
            
            if self._params["secondary_data_dist_transform_list"] is not None:
                f_dist_dg = self._params["secondary_data_dist_transform_list"][ii]
            else:
                f_dist_dg = None

            schema_info ("Flag 201.80 ", g_val.shape, g_type, gamma_g, f_dist_dg)
            
            if g_type == "categorical":
                dg = 1.0*( g_val[i_u] != g_val[i_v]) #1.0*( g_val[i_u].toarray() != g_val[i_v].toarray())
            elif g_type == "feature_vector":
                schema_info ("Flag 201.82 ", g_val[i_u].shape, g_val[i_v].shape)
                dg = np.ravel(np.sum(np.power(g_val[i_u].astype(np.float64) - g_val[i_v].astype(np.float64),2), axis=1))
            elif g_type == "feature_vector_categorical":
                schema_info ("Flag 201.84 ", g_val[i_u].shape, g_val[i_v].shape)
                dg = np.ravel(np.sum((g_val[i_u] == g_val[i_v]).astype(np.float64), axis=1))
            else:  #numeric
                dg = (g_val[i_u].astype(np.float64) - g_val[i_v].astype(np.float64))**2   #(g_val[i_u].toarray() - g_val[i_v].toarray())**2            

            if f_dist_dg is not None:
                dg = f_dist_dg(dg)
                
            z_dg = (dg - dg.mean())/dg.std()
            if gamma_g is not None:
                z_dg *= gamma_g
                
            l_z_dg.append(z_dg)

        schema_info ("Flag 201.99 ", uv_dist_centered.shape, z_d0.shape, len(l_z_dg), l_z_dg[0].shape)
        schema_info ("Flag 201.991 ", uv_dist_centered.mean(axis=0))
        schema_info ("Flag 201.992 ", uv_dist_centered.std(axis=0))
        schema_info ("Flag 201.993 ", z_d0[:10], z_d0.mean(), z_d0.std())
        schema_info ("Flag 201.994 ", l_z_dg[0][:10], l_z_dg[0].mean(), l_z_dg[0].std())
        return (uv_dist_centered, z_d0, l_z_dg)


    
    
    def _prepareQPterms(self, uv_dist_centered, z_d0, l_z_dg):
        nPointPairs = uv_dist_centered.shape[0]
        
        l_g = []
        for i,z_dg in enumerate(l_z_dg):
            l_g.append( np.sum(uv_dist_centered * z_dg[:,None], axis=0) )
            
        q1 = l_g[0]
        for v in l_g[1:]:
            q1 += v
        
        P1 = np.matmul( uv_dist_centered.T,  uv_dist_centered)

        g1 = np.sum(uv_dist_centered * z_d0[:,None], axis=0)
        h1 = np.sum(g1)

        return (P1, q1, g1, h1, nPointPairs)

    

    def _computeSolutionFeatures(self, w, P1, q1, g1, nPointPairs):
        K = len(w)
        
        newmetric_sd = np.sqrt( np.matmul(np.matmul(np.reshape(w,(1,K)), P1), np.reshape(w,(K,1)))[0] / nPointPairs)
        oldnew_corr = ((np.dot(w,g1)/nPointPairs)/newmetric_sd) / (self._params["d0_orig_transformed_corr"])

        groupcorr_score = (np.dot(w,q1)/nPointPairs)/newmetric_sd
        return {"w": w, "distcorr": oldnew_corr[0], "objval": groupcorr_score[0]}
        

    
    def _doQPSolve(self, P1, q1, g1, h1, nPointPairs, lambda1, alpha, beta):
        #https://cvxopt.org/examples/tutorial/qp.html and https://courses.csail.mit.edu/6.867/wiki/images/a/a7/Qp-cvxopt.pdf
        #  the cvx example switches P & q to Q & p
        
        from cvxopt import matrix, solvers
        solvers.options["show_progress"] = False


        K = len(g1)

        I_K = np.diagflat(np.ones(K).astype(np.float64))
        #P = 2* (lambda1*matrix(I_K) + beta*matrix(P1))
        P = matrix(2* ((I_K*lambda1) + (P1*beta)))
        
        q = -matrix(q1 + 2*lambda1*np.ones(K))
        
        G0 = np.zeros((K+1,K)).astype(np.float64)
        for i in range(K):
            G0[i,i] = 1.0
        G0[-1,:] = g1

                
        G = -matrix(G0) #first K rows say -w_i <= 0. The K+1'th row says -corr(newdist,olddist) <= -const
        h = matrix(np.zeros(K+1))
        h[-1] = -alpha*h1
        
        
        schema_debug("Flag 543.70 ", P1.shape, q1.shape, g1.shape, h1, nPointPairs, lambda1 , alpha, beta, P.size, q.size, G0.shape, G.size, h.size) #K, P.size, q.size, G.size, h.size)
        sol=solvers.qp(P, q, G, h)
        solvers.options["show_progress"] = True

        w = np.array(sol['x']).ravel()
        s = self._computeSolutionFeatures(w, P1, q1, g1, nPointPairs)
        return s

    
    def _summarizeSoln(self, soln, free_params):
        retl = []
        retl.append(("w_max_to_avg", (max(soln["w"])/np.nanmean(soln["w"]))))
        retl.append(("w_num_zero_dims", sum(soln["w"] < 1e-5)))
        retl.append(("d0_orig_transformed_corr", self._params["d0_orig_transformed_corr"]))
        retl.append(("distcorr", soln["distcorr"]))
        retl.append(("objval", soln["objval"]))
        retl.append(("lambda", free_params["lambda"]))
        retl.append(("alpha", free_params["alpha"]))
        retl.append(("beta", free_params["beta"]))
        
        return retl
                    

    
    def _doQPiterations(self, P1, q1, g1, h1, nPointPairs, max_w_wt, min_desired_oldnew_corr):
        solutionList = []

        schema_info('Progress bar (each dot is 10%): ', end='', flush=True)

        alpha = 1.0  #start from one (i.e. no limit on numerator and make it bigger)
        while alpha > 1e-5:
            soln, param_settings = self._iterateQPLevel1(P1, q1, g1, h1, nPointPairs, max_w_wt, min_desired_oldnew_corr, alpha)
            
            if soln is not None and soln["distcorr"] >= min_desired_oldnew_corr:
                solutionList.append((-soln["objval"], soln, param_settings))

            alpha -= 0.1
            schema_info('.', end='', flush=True)
        try:
            solutionList.sort(key=lambda v: v[0]) #find the highest score                

            schema_info(' Done\n', end='', flush=True)
            return (solutionList[0][1], solutionList[0][2])
        except:
            schema_info(' Done\n', end='', flush=True)
            #raise
            return (None, {})

        

        
    def _iterateQPLevel1(self, P1, q1, g1, h1, nPointPairs, max_w_wt, min_desired_oldnew_corr, alpha):
        solutionList = []

        beta = 1e6  #start from a large value and go towards zero (a large value means the denominator will need to be small)
        while beta > 1e-6:
            try:
                soln, param_settings = self._iterateQPLevel2(P1, q1, g1, h1, nPointPairs, max_w_wt, alpha, beta)
            except Exception as e:
                schema_info ("Flag 110.50 crashed in _iterateQPLevel2. Trying to continue...", P1.size, q1.size, g1.size, max_w_wt, alpha, beta)
                schema_info(e)
                beta *= 0.5
                continue
            
            if soln["distcorr"] >= min_desired_oldnew_corr:
                solutionList.append((-soln["objval"], soln, param_settings))

            beta *= 0.5
        
        try:
            solutionList.sort(key=lambda v: v[0]) #find the highest score

            schema_debug ("Flag 110.60 beta: ", "NONE" if not solutionList else self._summarizeSoln(solutionList[0][1], solutionList[0][2]))
            return (solutionList[0][1], solutionList[0][2])
        except:
            #raise
            return (None, {})



    def _iterateQPLevel2(self, P1, q1, g1, h1, nPointPairs, max_w_wt, alpha, beta):
        lo=1e-10 #1e-6
        solLo = self._doQPSolve(P1, q1, g1, h1, nPointPairs, 1/lo, alpha, beta)
        scalerangeLo = (max(solLo["w"])/np.nanmean(solLo["w"]))
        
        hi=1e9 #1e6
        solHi = self._doQPSolve(P1, q1, g1, h1, nPointPairs, 1/hi, alpha, beta)
        scalerangeHi = (max(solHi["w"])/np.nanmean(solHi["w"]))

        if scalerangeHi < max_w_wt:
            solZero = self._doQPSolve(P1, q1, g1, h1, nPointPairs, 0, alpha, beta)
            scalerangeZero = (max(solZero["w"])/np.nanmean(solZero["w"]))
            if scalerangeZero < max_w_wt and int(self._params.get("require_nonzero_lambda",0))==0:
                return solZero, {"lambda": 0, "alpha": alpha, "beta": beta} 
            else:
                return solHi, {"lambda": 1/hi, "alpha": alpha, "beta": beta}
            
        if scalerangeLo > max_w_wt: return solLo, {"lambda": 1/lo, "alpha": alpha, "beta": beta}
        
        niter=0
        while (niter < 60 and (hi/lo -1)>0.001):
            mid = np.exp((np.log(lo)+np.log(hi))/2)
            solMid = self._doQPSolve(P1, q1, g1, h1, nPointPairs, 1/mid, alpha, beta)
            scalerangeMid = (max(solMid["w"])/np.nanmean(solMid["w"]))
            
            if (scalerangeLo <= max_w_wt <= scalerangeMid):
                hi = mid
                scalerangeHi = scalerangeMid
                solHi = solMid
            else:
                lo = mid
                scalerangeLo = scalerangeMid
                solLo = solMid

            niter += 1
            schema_debug ("Flag 42.113 ", niter, lo, mid, hi, max(solLo["w"]), min(solLo["w"]), max(solHi["w"]), min(solHi["w"]), scalerangeMid)
        return solLo, {"lambda": 1/lo, "alpha": alpha, "beta": beta}        

    

    def _fit_helper(self, dx, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list):
        nPointPairs = self._params.get("dist_npairs", 2000000)
        sample_rate = self._params.get("dist_df_frac", 1.0)

        w_max_to_avg= self._w_max_to_avg
        min_desired_corr = self._min_desired_corr

        N = dx.shape[0]

        schema_info ("Flag 102.30 ", dx.shape, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
        
        G = []
        for i in range(len(secondary_data_val_list)):
            G.append((secondary_data_val_list[i].copy(), secondary_data_type_list[i], None if secondary_data_wt_list is None else secondary_data_wt_list[i]))

        if sample_rate < 0.9999999: 
            idx = np.random.choice(N, size=int(sample_rate*N), replace=False)
            dx = dx[idx,:]
            if d0 is not None:
                d0 = d0[idx,:]
            for i, gx in enumerate(G):
                G[i] = (gx[0][idx], gx[1], gx[2])


        schema_info ("Flag 102.35 ", dx.shape, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
        uv_dist_centered, z_d0, l_z_dg = self._getDistances(dx, d0, G, nPointPairs)
        schema_info ("Flag 102.36 ", uv_dist_centered.shape, z_d0.shape, len(l_z_dg))
        P1, q1, g1, h1, nPointPairs1 = self._prepareQPterms(uv_dist_centered, z_d0, l_z_dg)
        schema_info ("Flag 102.37 ")
        
        soln, free_params = self._doQPiterations(P1, q1, g1, h1, nPointPairs1, w_max_to_avg, min_desired_corr)
        
        if soln is None:
            raise Exception("Couldn't find valid solution to QP")
        
        schema_info ("Flag 102.40 Final solution: ", self._summarizeSoln(soln, free_params))
        return soln["w"].ravel(), self._summarizeSoln(soln, free_params)

    
        
    def _fit_scale(self, d, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list):

        dx1 = d.copy()
        if self._params.get("scale_mode_uses_standard_scaler",0)>0:
            self._std_scaler = sklearn.preprocessing.StandardScaler() 
            dx = self._std_scaler.fit_transform(dx1)
        else:
            self._std_scaler = sklearn.preprocessing.StandardScaler(with_mean=False, with_std=False) 
            dx = self._std_scaler.fit_transform(dx1) #this is a little wasteful no-op but makes a code a little easier to manage
            
        print('Running quadratic program...', end='', flush=True)
        s, sl = self._fit_helper(dx, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
        
        self._wts = np.maximum(s,0) # shouldn't be <0 anyways, except for numerical issues
        self._wts = len(self._wts)*self._wts/np.sum(self._wts) #let's normalize so that avg wt = 1
        
        self._soln_info = dict(sl)
        print(' done.\n', end='', flush=True)
                

        
    def _fit_affine(self, d, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list):
        do_whiten = self._params.get("do_whiten",1)>0
        ncomp = self._params.get("num_top_components",None) #default is all
        model_type = self._params.get("decomposition_model","pca").lower()
        dx1 = d.copy()
        
        if self._decomp_mdl is None:
            print('Running change-of-basis transform ({0}, {1} components)...'.format(model_type, ncomp), end='', flush=True)

            if model_type=="pca":
                if not scipy.sparse.issparse(dx1):
                    self._decomp_mdl = sklearn.decomposition.PCA(n_components=ncomp, whiten=do_whiten)
                else:
                    self._decomp_mdl = sklearn.decomposition.TruncatedSVD(n_components=ncomp)
                    schema_warning("Using TruncatedSVD instead of PCA because input is a sparse matrix. do_whiten arguments will be ignored")
                    
                self._decomp_mdl.fit(dx1)

            elif model_type=="nmf":
                self._decomp_mdl = sklearn.decomposition.NMF(n_components=ncomp, init=None)
                W = self._decomp_mdl.fit(dx1)
                
            self._orig_decomp_mdl = copy.deepcopy(self._decomp_mdl)
            print(' done.')
            
        else:
            print('Reusing previous change-of-basis transformation')
            self._decomp_mdl = copy.deepcopy(self._orig_decomp_mdl)

        if model_type=="pca":
            dx = self._decomp_mdl.transform(dx1)

        elif model_type=="nmf":
            W = self._decomp_mdl.transform(dx1)
            if do_whiten:
                H = self._decomp_mdl.components_
                wsd = W.std(axis=0)
                W = W/wsd
                self._decomp_mdl.components_ *= wsd[:,None]
            dx = W


            
        print('Running quadratic program...', end='', flush=True)
        s, sl = self._fit_helper(dx, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)

        self._wts = np.maximum(s,0) # shouldn't be <0 anyways, except for numerical issues
        self._wts = len(self._wts)*self._wts/np.sum(self._wts) #let's normalize so that avg wt = 1

        self._soln_info = dict(sl)
        print(' done.\n', end='', flush=True)





from setuptools import setup

setup(name='schema_learn',
      version='0.0.71',
      description='Metric learning enables synthesis of heterogeneous single-cell modalities',
      url='http://github.com/rs239/schema',
      author='Rohit Singh',
      author_email='rsingh@alum.mit.edu',
      license='MIT',
      packages=['schema'],
      install_requires = 'numpy,scipy,pandas,sklearn,cvxopt,tqdm'.split(','),
      zip_safe=False)
